"""
    CdInspection.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.ticker as ticker
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from OurPlotUtils import draw_as_image
from DataUtils import cut_upper_folders
from SvdDenoise import get_denoised_data
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
from molass_legacy.KekLib.SciPyCookbook import smooth
from .ConcDepend import compute_min_norm_scaled, compute_min_norm_bq_s_aq

@ticker.FuncFormatter
def major_formatter(x, pos):
    apparant = False if pos is None else (pos - 1)%2 == 0
    return "%.0f" % x if apparant else ""

class CdInspectionDailog(Dialog):
    def __init__( self, parent, dialog, M, E, C, q, eslice, from_ax=None, xray_scale=None, plotter_info=None):
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.dialog = dialog
        self.M = M
        self.E = E
        self.svd = np.linalg.svd(M)
        self.C = C
        self.q = q
        self.xray_picking = get_xray_picking()
        self.q_index = bisect_right(q, self.xray_picking)
        self.q_index2 = bisect_right(q, self.xray_picking+0.01)
        self.q_limit = bisect_right(q, 0.2)

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from MatrixData import simple_plot_3d
            print('M.shape=', M.shape, 'xray_scale=', xray_scale, 'q_limit=', self.q_limit, 'sigmas=', self.svd[1][0:2])
            c = C[0,:]
            c = c/np.max(c)
            plt.push()
            fig = plt.figure(figsize=(14,7))
            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122)
            simple_plot_3d(ax1, M)
            ax2.plot(c)
            fig.tight_layout()
            plt.show()
            plt.pop()

        self.eslice = eslice    # used just in title and logging
        self.plotter_info = plotter_info
        self.from_ax = from_ax
        self.xray_scale = xray_scale
        self.showing_sn = True
        self.popup_menu = None
        Dialog.__init__( self, parent, "Conc. Dependency Inspection", visible=False )

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig)
        # print("CdInspectionDailog: closed fig")
        Dialog.cancel(self)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(23,10))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        in_folder = cut_upper_folders(get_setting('in_folder'))
        fig.suptitle("Conc. Dependency Inspection on Elution Range %d-%d in %s" % (self.eslice.start, self.eslice.stop, in_folder), fontsize=28)

        self.create_axes(fig)
        self.compute_scores()
        self.draw_elution(self.ax0)
        self.draw_singular_values(self.ax1)

        self.draw_curves()

        fig.subplots_adjust(top=0.9, bottom=0.07, wspace=0.3, hspace=0.4, left=0.05, right=0.97)

        self.mpl_canvas.draw()

    def draw_curves(self):
        axes = self.axes

        score1, sy2 = self.draw_diff_curve(axes[2,0], self.A1, self.A2, "$A_1 - scale*A_2$", self.scale)
        self.draw_cd1_curve(axes[0,0], self.A1, "$A_1$", sy2=sy2, curve_label2="$scale*A_2$")
        self.draw_cd2_curve(axes[1,0], self.A2, "$A_2$", yb=self.B2)

        score2, sy2 = self.draw_diff_curve(axes[2,1], self.AE1, self.AE2, "$A_1/E_1 - scale*A_2/E_2$", self.scale2)
        self.draw_cd1_curve(axes[0,1], self.AE1, "$A_1/E_1$", sy2=sy2, curve_label2="$scale*A_2/E_2$")
        self.draw_cd2_curve(axes[1,1], self.AE2, "$A_2/E_2$", yb=self.B2)

        score3, sy2, sy2_no_bnd, score3_smoothed = self.draw_diff_curve(axes[2,2], self.A2, self.B2, "$B_2 - scale*A_2$", self.scale, aq_scale=False, with_smoothed=True)
        self.draw_cd1_curve(axes[0,2], self.A2, "$A_2$", log_scale=False, curve_label2="$scale*B_2$")
        self.draw_cd2_curve(axes[1,2], self.B2, "$B_2$", log_scale=False, sy2=sy2, sy2_no_bnd=sy2_no_bnd)

        sigmas = self.svd[1]
        self.logger.info("cd_scores=[%.3g, %.3g, %.3g, %.3g], sigmas=[%.3g, %.3g] for elution range(%d, %d)", score1, score2, score3, score3_smoothed, sigmas[0], sigmas[1], self.eslice.start, self.eslice.stop)

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(bottom_frame, width=width*0.25)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def create_axes(self, fig):
        gs = GridSpec(3,4)
        self.ax0 = fig.add_subplot(gs[0,0])
        self.ax1 = fig.add_subplot(gs[1:3,0])

        axes = []
        for i in range(3):
            ax_row = []
            for j in range(1,4):
                ax_row.append(fig.add_subplot(gs[i,j]))
            axes.append(ax_row)

        self.axes = axes = np.array(axes)

    def compute_scores(self):
        M1 = get_denoised_data(self.M, rank=1, svd=self.svd)
        C1 = self.C[0,:]
        C1 = C1/np.max(C1)
        C1inv = np.linalg.pinv(C1[np.newaxis,:])
        P1 = M1 @ C1inv
        E1 = np.sqrt((self.E**2) @ (C1inv**2))

        self.A1 = P1[:,0]
        self.AE1 = P1[:,0]/E1[:,0]

        M2 = get_denoised_data(self.M, rank=2, svd=self.svd)
        C2 = np.array([C1, C1**2])

        C2inv = np.linalg.pinv(C2)
        P2 = M2 @ C2inv
        E2 = np.sqrt((self.E**2) @ (C2inv**2))

        self.A2 = P2[:,0]
        self.AE2 = P2[:,0]/E2[:,0]
        self.B2 = P2[:,1]

        i_slice = slice(self.q_index, self.q_index2)
        if self.xray_scale is None:
            s1 = np.average(self.A1[i_slice])
            s2 = np.average(self.A2[i_slice])
            self.scale = np.sqrt(s1*s2)
        else:
            self.scale = self.xray_scale
        s1 = np.average(self.AE1[i_slice])
        s2 = np.average(self.AE2[i_slice])
        self.scale2 = np.sqrt(s1*s2)

    def draw_elution(self, ax):
        from_ax = self.from_ax
        if from_ax is None:
            plotter, i, pno = self.plotter_info
            self.logger.info("draw_elution: i=%d, pno=%d", i, pno)
            if i == 1:
                plotter.draw_fig2(ax, title=False, peak_no=pno)
            else:
                plotter.draw_mapped(ax, title=False, peak_no=pno)
        else:
            ax.set_axis_off()
            draw_as_image(ax, self.dialog.fig, from_ax, exp_elements=True)

    def draw_singular_values(self, ax):
        ax.set_title("Top Five Singular Values", fontsize=16)
        ax.set_xlabel("Value No", fontsize=12)
        ax.set_ylabel("Scale", fontsize=12)
        ax.xaxis.set_major_formatter(major_formatter)

        sigmas = self.svd[1]
        n = min(5, len(sigmas))
        """
            for cases where len(sigmas) < 5 as in 20190315_2
        """
        ax.plot(np.arange(n), sigmas[0:n], ':', marker='o')

    def draw_cd1_curve(self, ax, y, curve_label, log_scale=True, sy2=None, curve_label2=None):
        if log_scale:
            ax.set_yscale("log")
        ax.set_title("%s Curve" % curve_label, fontsize=16)

        ax.plot(self.q, y, color='C1', label=curve_label)
        if sy2 is not None:
            color = 'C2' if log_scale else 'pink'
            ax.plot(self.q, sy2, color=color, label=curve_label2, alpha=0.5)

        ax.legend()

    def draw_cd2_curve(self, ax, y, curve_label, yb=None, log_scale=True, sy2=None, sy2_no_bnd=None):
        if log_scale:
            ax.set_yscale("log")
        ax.set_title("%s Curve" % curve_label, fontsize=16)

        color = 'C2' if log_scale else 'pink'
        ax.plot(self.q, y, color=color, label=curve_label)

        if yb is not None:
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(self.q, yb, color='pink', label='$B_2$')
            axt.legend(bbox_to_anchor=(1, 0.82), loc='upper right')

            # delete ticks which appear due to twinx()
            ax.tick_params(which='both', left=False)

        if sy2 is not None:
            ax.plot(self.q, sy2, color='C1', label='$scale*A_2$')

        if sy2_no_bnd is not None:
            ax.plot(self.q, sy2_no_bnd, color='yellow', label='$scale*A_2$ (no bound)')

        ax.legend()

    def draw_diff_curve(self, ax, y1, y2, curve_label, scale, aq_scale=True, with_smoothed=True):
        if aq_scale:
            sy2 = compute_min_norm_scaled(y1, y2)
            y = y1 - sy2
        else:
            sigmas = self.svd[1]
            init_scale = sigmas[1]/sigmas[0]
            sy2, sy2_no_bnd = compute_min_norm_bq_s_aq(y2, y1, init_scale, self.logger, ret_no_bnd=True)
            y = y2 - sy2

        ax.set_title("Scaled Difference Curve", fontsize=16)

        aslice = slice(0,self.q_limit)
        y_ = y[aslice]
        cd_score = np.sqrt(np.average(y_**2))*100/scale

        if not aq_scale and sy2_no_bnd is not None:
            y = y2 - sy2_no_bnd
            ax.plot(self.q, y, color='yellow', label=curve_label + ' (no bound)')
            y_ = y[aslice]
            cd_score_no_bnd = np.sqrt(np.average(y_**2))*100/scale
        else:
            cd_score_no_bnd = None

        if aq_scale:
            ax.plot(self.q, y, color='gray', label=curve_label)
            self.draw_score_text(ax, y[-1], cd_score, cd_score_no_bnd)
        else:
            if with_smoothed:
                b = smooth(y2)
                a = smooth(y1)
                sa, min_scale = compute_min_norm_bq_s_aq(b, a, init_scale, ret_scale=True)
                y_ = b - sa
                cd_score_smoothed = np.sqrt(np.average(y_[aslice]**2))*100/scale
                ax.plot(self.q, y_, color='blue', label='$smooth\ B_2 - scale* smooth\ A_2$')

                ax.plot(self.q, y, color='cyan', alpha=0.5, label=curve_label)
                self.draw_score_text(ax, y[-1], cd_score_smoothed, cd_score, min_scale=min_scale)
            else:
                ax.plot(self.q, y, color=color, label=curve_label)
                self.draw_score_text(ax, y[-1], cd_score, None)

        ax.legend()

        if aq_scale:
            return cd_score, sy2
        else:
            if with_smoothed:
                return cd_score, sy2, sy2_no_bnd, cd_score_smoothed
            else:
                return cd_score, sy2, sy2_no_bnd

    def draw_score_text(self, ax, yp, score, cd_score_no_bnd, min_scale=None):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = xmin*0.4 + xmax*0.6
        yratio = (yp - ymin)/(ymax - ymin)
        if yratio > 0.5:
            ty = ymin*0.4 + yp*0.6
        else:
            ty = yp*0.6 + ymax*0.4
        ty_ = ty if cd_score_no_bnd is None else ty + (ymax - ymin)*0.11
        ax.text(tx, ty_, "SCD=%.3g" % score, ha='center', va="center", alpha=0.5, fontsize=40, color='blue')
        if cd_score_no_bnd is not None:
            ty_ = ty - (ymax - ymin)*0.11
            ax.text(tx, ty_, "SCD=%.3g" % cd_score_no_bnd, ha='center', va="center", alpha=0.5, fontsize=40, color='cyan')
        if min_scale is not None:
            ty_ = ty - (ymax - ymin)*0.34
            ax.text(tx, ty_, "scale=%.3g" % min_scale, ha='center', va="center", alpha=0.5, fontsize=40, color='gray')


    def save_the_figure(self, folder, pno, ad=None):
        import os
        from molass_legacy._MOLASS.SerialSettings import get_setting
        analysis_name = get_setting('analysis_name')
        ad_ = '' if ad is None else '-%d' % ad
        filename = analysis_name.replace( 'analysis', 'figure' ) + '-%d' % pno + ad_
        path = os.path.join( folder, filename )
        self.fig.savefig( path )



"""
    SurveyUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

from molass_legacy._MOLASS.SerialSettings import get_setting

def write_info(survey_fh, mapper):
    in_folder = get_setting('in_folder')
    x = mapper.x_curve.x
    cx = x[len(x)//2]
    cx_uv = []
    A, B = mapper.map_params
    for a, b in [(mapper.A_init, mapper.B_init), (A, B)]:
        cx_uv.append(cx*a + b)
    diff = cx_uv[0] - cx_uv[1]
    survey_params = ["%g" % v for v in [mapper.A_init, mapper.B_init, A, B, diff, diff*len(x)/len(mapper.a_curve.x), *mapper.get_sci_list()]]
    survey_fh.write(",".join([in_folder] + survey_params) + "\n")
    survey_fh.flush()



# coding: utf-8
"""
    Synthesized.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgb
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from SvdDenoise import get_denoised_data
# import molass_legacy.KekLib.DebugPlot as plt
from DataUtils import get_in_folder
# from .Boundary import get_boundary
from Theory.SolidSphere import get_boundary_params

def solve(M, E, c, qv, b, k, debug=False):
    M2 = get_denoised_data(M, rank=2)
    M1 = get_denoised_data(M, rank=1)

    C2 = np.array([c, c**2])
    P2 = M2 @ np.linalg.pinv(C2)
    C1 = np.array([c])
    P1 = M2 @ np.linalg.pinv(C1)

    w = 1/(1+np.exp(-k*(qv-b)))

    M_ = (M2.T*(1-w) + M1.T*w).T

    P_ = M_ @ np.linalg.pinv(C2)
    py_list = []
    py_list.append(P_[:,0].copy())

    for i in range(0):
        P_[b:,1] = 0
        C_ = np.linalg.pinv(P_) @ M_
        P_ = M_ @ np.linalg.pinv(C_)
        py_list.append(P_[:,0].copy())

    return P1, P2, w, py_list

class Demo(Dialog):
    def __init__(self, parent, *args):
        self.args = args
        self.fig = None
        self.almerge_result = None
        Dialog.__init__(self, parent, "Synthesized LRF Demo", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig, axes = plt.subplots(ncols=3, figsize=(21,7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.fig = fig
        self.axes = axes
        self.ax2t = None
        self.show_bq = Tk.IntVar()
        self.show_bq.set(0)
        self.show_almerge = Tk.IntVar()
        self.show_almerge.set(0)

        self.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(bottom_frame, width=width*0.25)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)
        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        panel = Tk.Frame(bottom_frame)
        panel.pack(side=Tk.RIGHT, padx=padx)

        w = Tk.Checkbutton(panel, text="show ALMERGE info", variable=self.show_almerge)
        w.pack(side=Tk.LEFT, padx=20)

        w = Tk.Checkbutton(panel, text="show B(q)", variable=self.show_bq)
        w.pack(side=Tk.LEFT, padx=20)

        w = Tk.Button(panel, text="Redraw", width=10, command=self.draw)
        w.pack(side=Tk.LEFT, padx=20)

    def draw(self):
        qv, P1, P2, b, w, py_list = self.args[3:]

        fig = self.fig
        for ax in self.axes:
            ax.cla()
        ax1, ax2, ax3 = self.axes
        if self.show_bq.get():
            if self.ax2t is None:
                ax2t = ax2.twinx()
                self.ax2t = ax2t
            else:
                ax2t = self.ax2t
            ax2t.cla()
            ax2t.grid(False)
        else:
            if self.ax2t is not None:
                self.ax2t.remove()
                self.ax2t = None

        fig.suptitle("Logistically Synthesized LRF Demo for " + get_in_folder(), fontsize=20)
        ax1.set_title("Rank(1,1) Weights (Logistic Function)", fontsize=16)
        ax2.set_title("Separate LRF", fontsize=16)
        ax3.set_title("Logistically Synthesized LRF", fontsize=16)

        ax1.set_ylabel('Weight')
        ax1.set_xlabel('Q')

        for ax in [ax2, ax3]:
            ax.set_yscale('log')
            ax.set_ylabel('$Log_{10}(I)$')
            ax.set_xlabel('Q')

        c0 = np.ones((len(w),3))*to_rgb('C0')
        c1 = np.ones((len(w),3))*to_rgb('C1')
        c_ = (c0.T*(1-w) + c1.T*w).T

        for i in range(len(w)):
            ax1.plot(qv[i:i+2], w[i:i+2], color=c_[i])

        for rank, y in [(2, P2[:,0]), (1, P1[:,0])]:
            ax2.plot(qv, y, label='rank({0},{0})'.format(rank), alpha=0.5)

        if self.show_bq.get():
            ax2t.plot(qv, P2[:,1], color='pink', label='B(q) in Rank(2,2)')

        y = py_list[-1]
        for i in range(len(w)):
            ax3.plot(qv[i:i+2], y[i:i+2], color=c_[i])

        ax3.set_ylim(ax2.get_ylim())

        bq = b
        for ax in [ax1, ax2, ax3]:
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            ax.plot([bq, bq], [ymin, ymax], ':', color='red', label='rank boundary')

        if self.show_almerge.get():
            abq, ay = self.get_almerge_results()
            ax2.plot(qv, ay, label='almerge extrapolated', alpha=0.5)
            ymin, ymax = ax2.get_ylim()
            ax2.plot([abq, abq], [ymin, ymax], color='red', label='almerge boundary')

        ax1.legend()
        ax2.legend()
        if self.show_bq.get():
            ax2t.legend(bbox_to_anchor=(1, 0.85), loc='upper right')
        ax3.legend()
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        # plt.show()
        self.mpl_canvas.draw()

    def get_almerge_results(self):
        abq, ay = self.compute_almerge_results()
        return abq, ay

    def compute_almerge_results(self):
        from SerialAtsasTools import AlmergeExecutor

        if self.almerge_result is None:
            almerge = AlmergeExecutor()
            M, E, c, qv = self.args[0:4]
            result = almerge.execute_matrix(qv, M, E, c)
            self.almerge_result = result
        else:
            qv = self.args[3]
            result = self.almerge_result
        i = result.overlap_from_max
        return qv[i], result.exz_array[:,1]

def demo(parent, sd, debug=True):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[0]
    f = range_[0]
    p = range_[1]
    t = range_[2]+1
    eslice = slice(f,t)
    M_ = M[:,eslice]
    E_ = E[:,eslice]

    Rg = sd.pre_recog.get_rg()
    b1, b2, k = get_boundary_params(Rg)

    x = ecurve.x
    y = ecurve.y
    c = y[f:t].copy()
    c /= np.max(c)

    P1, P2, w, py_list = solve(M_, E_, c, qv, b1, k, debug=debug)

    if debug:
        dialog = Demo(parent, M_, E_, c, qv, P1, P2, b1, w, py_list)
        dialog.show()

def get_reduced_rank(rank, cd):
    if cd == 1:
        ret_rank = rank
    else:
        ret_rank = rank - 1 if rank < 4 else rank - 2
    return ret_rank

def synthesized_data(qv, M, Rg, rank=2, cd=1, boundary=None, k=None, logger=None):
    M2 = get_denoised_data(M, rank=rank)
    rank_ = get_reduced_rank(rank, cd)
    M1 = get_denoised_data(M, rank=rank_)
    if boundary is None:
        assert k is None
        boundary, b2, k = get_boundary_params(Rg)
    else:
        assert k is not None

    w = 1/(1+np.exp(-k*(qv-boundary)))
    M_ = (M2.T*(1-w) + M1.T*w).T

    return M_, boundary



"""
    Optimizer.NumComponents.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.Optimizer.TheUtils import guess_n_compnents_trial
from molass_legacy.Models.ElutionCurveModels import EGHA

MAXNUM_COMPONENTS = 5

class NumComponents:
    def __init__(self, fullopt_input):
        self.logger = logging.getLogger(__name__)
        self.fullopt_input = fullopt_input
        self.dsets = fullopt_input.get_dsets()
        self.decomp_result = None
        self.n_components = None

    def get_decomp_result(self, fullopt_input, mapped_params):
        from DecompUtils import CorrectedBaseline, decompose_elution_better

        A, B, num_peaks_mapping = mapped_params
        sd = fullopt_input.sd
        mapper = fullopt_input.mapper

        self.logger.info("guessing init_params with v1decomposer")

        corbase_info = CorrectedBaseline(sd, mapper)
        decomp_result = decompose_elution_better(corbase_info, mapper, model=EGHA(), logger=self.logger)

        self.decomp_result = decomp_result
        return decomp_result

    def guess(self):
        if self.decomp_result is None:
            # this case is not yet tested
            mapped_params = self.fullopt_input.get_mapped_params()
            self.get_decomp_result(self.fullopt_input, mapped_params)

        # self.n_components = guess_n_compnents_trial(self.dsets, self.logger)
        self.n_components = self.guess_from_decomp_opt_recs()

    def get(self):
        if self.n_components is None:
            self.guess()
        return self.n_components

    def guess_from_decomp_opt_recs(self, debug=False):
        from DecompUtils import debug_elution_plot

        xr_curve, xrD = self.dsets[0]
        x = xr_curve.x
        y = xr_curve.y
        opt_recs = self.decomp_result.opt_recs

        if debug:
            debug_elution_plot("guess_from_decomp_opt_recs", EGHA(), x, y, opt_recs)

        nc = 0
        for k, rec in enumerate(opt_recs):
            ev = rec.evaluator
            h = ev.get_param_value(0)
            if debug:
                print([k], ev.sign, h)
            if ev.sign > 0:
                nc += 1

        return nc



"""
    Jupyter.JupyterLab.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import re

def run_jupyterlab(parent=None, try_install=True):
    if try_install:
        from .Installer import install_jupyterlab
        ok = install_jupyterlab(parent)
    else:
        ok = True

    if not ok:
       return

    import sys
    import os
    import subprocess

    this_dir = os.path.dirname(__file__)
    home_dir = os.path.abspath(this_dir + "/../../")
    notebooks_path_txt = os.path.join(home_dir, "notebooks-path.txt")

    with open(notebooks_path_txt) as fh:
        notebooks_path = re.sub(r"\s+", "", fh.read())
    # print("notebooks_path='%s'" % notebooks_path)
    if not os.path.exists(notebooks_path):
        from molass_legacy.KekLib.TkCustomWidgets import MessageBox
        # HookableMessageBox would be required if this must be really the custom widget.
        MessageBox.showinfo("Notebooks path notification",
            'The folder path "%s",\n'
            'which is written in "%s",\n'
            "does not exist.\n"
            "This folder will be created soon after this confirmation\n"
            "and will be used by Jupyter as the current folder.\n\n"
            "Please rewrite it and retry if you wish to change it."
            % (notebooks_path, notebooks_path_txt),
            parent=parent
            )
        os.makedirs(notebooks_path)

    os.chdir(notebooks_path)

    jupyterlab_exe = os.path.join(os.path.dirname(sys.executable), "Scripts\\jupyter-lab.exe")
    # note that "jupyter.exe lab" invocation insead of "jupyter-lab.exe"
    # would fail if the folder has been moved from the created location.

    # quotes are required for paths like "C:\Program Files\Python310\Scripts\jupyter-lab.exe".
    jupyterlab_exe_quoted = '"%s"' % jupyterlab_exe
    print("jupyterlab_exe='%s'" % jupyterlab_exe)
    os.spawnl(os.P_DETACH, jupyterlab_exe, jupyterlab_exe_quoted)   # run in background



# coding: utf-8
"""
    BiComponentSolver.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF

"""
from bisect                 import bisect_right
import numpy                as np
from scipy                  import optimize
from DataModelsAgg          import _GuinierPorodAgg
from DataModels             import _GuinierPorod

REQUIRED_BASIC_QUALITY  = 0.7

class BiComponentSolver:
    def __init__( self, intensity, no_assert=False ):

        if not no_assert:
            assert( intensity.basic_quality >=REQUIRED_BASIC_QUALITY )

        self.intensity  = intensity
        self.smoother   = intensity.smoother
        self.I0_est     = None
        self.start      = 0         # TODO: remove this

    def solve( self, fit=None ):
        from lmfit import minimize, Parameters

        assert( fit is not None )

        if fit is None:
            G1_init     = self.intensity.SQ[0]
            Rg1_init    = 50.0
            Rg2_init    = 20.0
            d1_init     = 3.0
            d2_init     = 3.0
        else:
            self.I0_est = fit.I0
            G1_init     = fit.I0 * 0.7
            Rg1_init    = fit.Rg * np.sqrt( 2.0 )
            Rg2_init    = Rg1_init / 2
            d1_init     = fit.degree
            d2_init     = d1_init

        Q1 = 1/Rg1_init * np.sqrt( 3*d1_init/2 )
        Q2 = 1/Rg2_init * np.sqrt( 3*d2_init/2 )

        self.prepare_equation_data( Q1, Q2 )

        def f(p):
            return np.abs( np.sum( np.array( self.equations( p ) )**2 ) - 0 )

        min_ee      = None
        min_params  = None

        for ratio in [ 0.5, 0.6, 0.7, 0.8, 0.9 ]:

            G1_init_ = self.I0_est * ratio
            p_init = optimize.fmin( f, ( G1_init_, Rg1_init, Rg2_init, d1_init, d2_init ), disp=False )

            G1, Rg1, Rg2, d1, d2 = optimize.fsolve( self.equations, p_init )
            G2 = self.compute_G2( G1, Rg1, Rg2 )

            Q1 = 1/Rg1 * np.sqrt( 3*d1/2 )
            Q2 = 1/Rg2 * np.sqrt( 3*d2/2 )

            self.prepare_equation_data( Q1, Q2 )
            p_init = optimize.fmin( f, ( G1, Rg1, Rg2, d1, d2 ), disp=False )
            G1, Rg1, Rg2, d1, d2 = optimize.fsolve( self.equations, p_init )
            G2 = self.compute_G2( G1, Rg1, Rg2 )

            if Rg1 < Rg2:
                G1, G2, Rg1, Rg2, d1, d2 = G2, G1, Rg2, Rg1, d2, d1

            params  = Parameters()
            params.add('G1',    value= G1,   min=1e-5, max=self.I0_est )
            params.add('Rg1',   value= Rg1,  min=1.0,  max=500 )
            params.add('Rg2',   value= Rg2,  min=1.0,  max=500 )
            params.add('d1',    value= d1,   min=1.0,  max=5.0 )
            params.add('d2',    value= d2,   min=1.0,  max=5.0 )
            result = minimize( self.lmfit_model, params, args=() )

            G1  = result.params['G1'].value
            Rg1 = result.params['Rg1'].value
            Rg2 = result.params['Rg2'].value
            d1  = result.params['d1'].value
            d2  = result.params['d2'].value
            G2  = self.compute_G2( G1, Rg1, Rg2 )

            opt_params  = [ G1, G2, Rg1, Rg2, d1, d2 ]
            opt_ee      = self.evaluate( *opt_params )

            if min_ee is None or opt_ee < min_ee:
                min_ee      = opt_ee
                min_params  = opt_params

        # print( 'BiComponentSolver.solve: min_params=', min_params[2:4] )
        return [ min_ee, min_params, self.I0_est ]

    def prepare_equation_data( self, Q1, Q2 ):
        self.Q_array = np.array( [
                    0.25 * Q1,
                    0.75 * Q1,
                    0.75 * Q1 + 0.25 * Q2,
                    0.25 * Q1 + 0.75 * Q2,
                    1.4  * Q2,
                    min( 2.0 * Q2, self.intensity.Q[-1]),
                  ] )
        self.I_array = self.smoother( self.Q_array )

    def equations( self, p ):
        G1, Rg1, Rg2, d1, d2 = p

        if not ( d1>=1 and d1 <= 5 and d2>=1 and d2 <= 5 ):
            return ( np.inf, ) * len(p)

        if G1 < 0:
            return ( np.inf, ) * len(p)

        G2 = self.compute_G2( G1, Rg1, Rg2 )
        if G2 < 0:
            return ( np.inf, ) * len(p)

        if Rg1 < Rg2:
            return ( np.inf, ) * len(p)

        qq = self.Q_array[1]**2
        f2 = G1 * np.exp( - qq * Rg1**2/3 ) + G2 * np.exp( - qq * Rg2**2/3 ) - self.I_array[1]

        D1 = G1 * np.exp( -d1/2 ) * np.power( 3*d1/2, d1/2 ) / np.power( Rg1, d1 )
        D2 = G2 * np.exp( -d2/2 ) * np.power( 3*d2/2, d2/2 ) / np.power( Rg2, d2 )

        q  = self.Q_array[2]
        f3 = D1 / q**d1 + G2 * np.exp( - q**2 * Rg2**2 / 3 ) - self.I_array[2]
        q  = self.Q_array[3]
        f4 = D1 / q**d1 + G2 * np.exp( - q**2 * Rg2**2 / 3 ) - self.I_array[3]

        q  = self.Q_array[4]
        f5 = D1 / q**d1 + D2 / q**d2 - self.I_array[4]
        q  = self.Q_array[5]
        f6 = D1 / q**d1 + D2 / q**d2 - self.I_array[5]

        return ( f2, f3, f4, f5, f6 )
        # この戻り値が ( 0, 0, 0, 0, 0 ) になるような G1, Rg1, Rg2, d1, d2
        # を求める。

    def compute_G2( self, G1, Rg1, Rg2 ):
        if self.I0_est is None:
            qq = self.intensity.SQ[0]**2
            G2 = ( self.intensity.SI[0] - G1 * np.exp( - qq * Rg1**2/3 ) ) / np.exp( - qq * Rg2**2/3 )
        else:
            G2 = self.I0_est - G1
        return G2

    def evaluate( self, G1, G2, Rg1, Rg2, d1, d2, debug=False, caller='unknown', eval_max=None ):

        Y_  = np.log( _GuinierPorodAgg( [G1, G2], [Rg1, Rg2], [d1, d2], self.intensity.SQ ) )

        e = self.intensity.SW * ( self.intensity.SY - Y_ )
        ee = np.inner( e, e ) / ( self.intensity.SQ.shape[0] - 3 )

        if ( debug
            and self.debugger_queues is not None
            and ( eval_max is None or ee < eval_max )
            ):
            print( 'evaluate: debugger_queues get' )
            dbcmd = self.debugger_queues[0].get( block=True )
            print( 'evaluate: dbcmd=', dbcmd )
            if dbcmd == 'Stop':
                self.debug_stop = True
                raise Exception( 'Stop' )
            elif dbcmd == 'Step':
                self.debugger_queues[1].put( [ caller, ee, [ G1, G2, Rg1, Rg2, d1, d2 ] ] )
            else:
                assert( False )

        return ee

    def evaluate_gp( self, I0, Rg, d ):
        """
            Be aware that this evaluation is performed on the same interval as
            that of the final solution of the solver.
        """
        Y_ = np.log( _GuinierPorod( I0, Rg, d, self.intensity.SQ ) )
        e = self.intensity.SW * ( self.intensity.SY - Y_ )
        ee = np.inner( e, e ) / ( self.intensity.SQ.shape[0] - 3 )

        return ee

    def lmfit_model( self, params ):
        G1  = params['G1'].value
        Rg1 = params['Rg1'].value
        Rg2 = params['Rg2'].value
        d1  = params['d1'].value
        d2  = params['d2'].value

        G2  = self.compute_G2( G1, Rg1, Rg2 )

        Y_  = np.log( _GuinierPorodAgg( [G1, G2], [Rg1, Rg2], [d1, d2], self.intensity.SQ ) )
        e = self.intensity.SW * ( self.intensity.SY - Y_ )
        return e



# coding: utf-8
"""
    SvdPrecision.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.OurMatplotlib import get_color
from ModeledData import ModeledData, simple_plot_3d
from XrayData import XrayData
from molass_legacy.ElutionDecomposer import ElutionDecomposer
from ExistingNmf import ExistingNmf

def svd_variation_by_noise():
    qvector = np.linspace(0.01, 0.6, 600)
    rg_list = [50, 24]
    d_list = [1, 3]
    md = ModeledData(qvector, 300, rg_list=rg_list, d_list=d_list)
    md.plot_components()
    M = md.get_data()
    fig = plt.figure(figsize=(40,24))
    case_type = 'Weak' if d_list[0] == d_list[1] else 'Strong'
    fig.suptitle('Matrix Factorization Comparison in case of %s Linear Independence' % case_type, fontsize=40)
    title_fontsize = 16
    gs = GridSpec(8,5)
    axes1 = []
    axes2 = []
    E = np.ones(M.shape)
    mu = 0
    sigma = 1
    for j in range(5):
        ax1 = fig.add_subplot(gs[0,j], projection='3d')
        ax2 = fig.add_subplot(gs[1,j])
        ax3 = fig.add_subplot(gs[2,j])
        ax4 = fig.add_subplot(gs[3,j])
        ax5 = fig.add_subplot(gs[4,j])
        ax6 = fig.add_subplot(gs[5,j])
        ax7 = fig.add_subplot(gs[6,j])
        ax8 = fig.add_subplot(gs[7,j])

        noise = j*0.05
        ax1.set_title('Noise %d %%' % (int(noise*100)), fontsize=title_fontsize)
        ax2.set_title('SVD major U spectra', fontsize=title_fontsize)
        ax3.set_title('SVD major Sigular Values', fontsize=title_fontsize)
        ax5.set_title('Elution Curve Decomposition (SA)', fontsize=title_fontsize)
        ax6.set_title('Elution Curve Decomposition (PMF)', fontsize=title_fontsize)
        ax7.set_title('Component Scattering Curves (SA)', fontsize=title_fontsize)
        ax8.set_title('Component Scattering Curves (PMF)', fontsize=title_fontsize)

        axes1.append(ax1)
        axes2.append(ax2)
        E_ = E + np.random.normal(mu, sigma, M.shape) * noise
        M_ = M * E_
        simple_plot_3d(ax1, M_)
        U, s, VT = np.linalg.svd(M_)
        V = VT.T
        for k in range(2):
            ax2.plot(U[:,k])
        ax3.plot(s[0:5], ':')
        ax3.plot(s[0:5], 'o')
        for k in range(2):
            ax4.plot(V[:,k])
        ax1.set_zlim(-0.1, 1.2)
        ax3.set_ylim(-3, 65)
        xd = XrayData(None, q=qvector, data=M_, error=E_)
        e_curve = xd.e_curve
        x = e_curve.x
        y = e_curve.y
        ax5.plot(y)
        decomp = ElutionDecomposer(e_curve, x, y)
        c_list = []
        for rec in decomp.fit_recs:
            f = rec.evaluator
            c = f(x)
            ax5.plot(x, c)
            c_list.append(c)

        ax6.plot(y)
        nmf = ExistingNmf(M_)
        H = nmf.H
        W = nmf.W

        h1 = H[0,:]
        h2 = H[1,:]
        w1 = W[:,0]
        w2 = W[:,1]

        i = int(e_curve.peak_top_x[0])
        numbered_order = h1[i] > h2[i]
        h_list = [h1, h2] if numbered_order else [h2, h1]
        w_list = [w1, w2] if numbered_order else [w2, w1]

        scales = solve_scale(y, h_list)
        for k, h in enumerate(h_list):
            ax6.plot(x, h*scales[k])

        C_ = np.array(c_list)
        rank = 2
        Us_ = np.dot( U[:,0:rank], np.diag( s[0:rank] ) )
        MD  = np.dot( Us_, VT[0:rank,:] )
        Cinv = np.linalg.pinv(C_)
        P_ = np.dot(MD, Cinv)

        for k in range(P_.shape[1]):
            color = get_color(k+1)
            ax7.plot(qvector, md.P[:, k], ':', color=color, label='model truth[%d]' % k)
            ax7.plot(qvector, P_[:,k], color=color, label='decomposed[%d]' % k)
            w = w_list[k]
            ax8.plot(qvector, md.P[:,k], ':', color=color, label='model truth[%d]' % k)
            scale = 1/scales[k]
            ax8.plot(qvector, w*scale, color=color, label='decomposed[%d]' % k)

        for ax in [ax7, ax8]:
            ax.legend(fontsize=16)

        plot_text(ax2, 'u0, u1')
        plot_text(ax3, 's0..s4')
        plot_text(ax4, 'v0, v1')
        plot_text(ax5, 'SA')
        plot_text(ax6, 'PMF')
        plot_text(ax7, 'SA')
        plot_text(ax8, 'PMF')

    fig.tight_layout()
    fig.subplots_adjust(top=0.95)
    plt.show()

def solve_scale(y, h_list):
    h_array = np.array(h_list)

    def obj_func(scale):
        debug = False
        if debug:
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.set_title('scale=%.3g'% scale[0])
            ax.plot(y)

        y_ = np.zeros(len(y))
        for k in range(h_array.shape[0]):
            h = h_array[k,:]*scale[k]
            if debug:
                ax.plot(h)
            y_ += h

        if debug:
            plt.show()
            plt.pop()

        return np.sum((y - y_)**2)

    init_x = np.ones((2,))
    result = minimize(obj_func, init_x)
    return result.x

def plot_text(ax, text, alpha=0.05, fontsize=100):
    x = ax.get_xlim()
    y = ax.get_ylim()
    tx = np.average(x)
    ty = np.average(y)
    ax.text(tx, ty, text, alpha=alpha, fontsize=fontsize, ha='center', va='center')



#!/usr/bin/env python

# Calculate g(r) from a trajectory PDB file
# Chris Cioce, 2012, ccioce@mail.usf.edu
# Original script by Stas Bevc, 2011, stas.bevc@cmm.ki.si, www.sicmm.org

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

"""
BACKGROUND:
This script will calculate the 3D radial distribution function from a specified trajectory PDB file.
Trajectory PDB files are individual PDB files strung together, separated by the keyword "ENDMDL".
It is important for viewers, such as VMD (http://www.ks.uiuc.edu/Research/vmd/), to ensure that 
each of the configurations (idividual frames) has the same number of atoms. This can be problematic 
for viewing trajectories of Monte Carlo simulations. There are, however, ways of tricking the viewers 
to play the files. A perl script which does this can be downloaded from the MPMC/tools/ directory 
(called vmdfixn.pl) at http://code.google.com/p/mpmc/source/browse/#svn%2Ftrunk%2Ftools.

WHAT YOU NEED TO MODIFY IN THIS SCRIPT:
This script assumes that there are center of mass sites for every molecule in the trajectory PDB file.
For example, if N2 is the species under consideration, there assumes that halfway between the nitrogens
resides a site with definitive coordinates, let's call it "N2G". For molecules such as CO2 and CH4,
this is not an issue, as the carbon atoms are the center of masses of the molecules. If you do not 
have these center of mass sites for your molecule(s) of interest, consider downloading the original 
python script (http://www.sicmm.org/~stas/rdf-script/) from which mine is a modified version of.
This original script has a flag (cm=True) which will auto-magically compute the COM sites for your
molecules.

This only one line which needs to be changed for you to successfully run this script, is the line

if line[12:15] != "N2G":                   # XXX Identifying label for center of gravity (COM) site

This line (line 67) is clearly marked with XXX, which highlights in editors such as vim.
"""

import sys
from os import listdir, getcwd
from math import sqrt, ceil, floor, pow, pi

# Compute g(r) from a trajectory PDB file
def calcRDF(trajfile, pdbdir, side, n_bins):

    print "Calculating g(r)..."
    traj = open(pdbdir+"/"+trajfile)

    maxbin = n_bins                                                                     # Number of requested bins
    box_half = side * 0.5
    dr = float(box_half/maxbin)                                                         # Bin width
    hist = [0]*(maxbin+1)
    rdf = {}
    config = 1                                                                          # Configuration (step) counter
    part_total = 0
    
    # Extract COM sites from trajectory PDB
    atoms = []
    for line in traj:                                                                   # Read entire trajectory file

        if line[12:15] != "N2G":                                                        # XXX Identifying label for center of gravity (COM) site

            if line[0:6] == "ENDMDL":                                                   # We've reached the end of a configuration
                # loop over particle pairs
                npart = len(atoms)
                part_total += npart
                for i in range(npart):
                        
                    xi = (atoms[i])[0]
                    yi = (atoms[i])[1]
                    zi = (atoms[i])[2]
                        
                    for j in range(i+1, npart):
                        xx = xi - (atoms[j])[0]
                        yy = yi - (atoms[j])[1]
                        zz = zi - (atoms[j])[2]
                            
                        # minimum image
                        if (xx < -box_half):   xx = xx + side
                        if (xx >  box_half):   xx = xx - side
                        if (yy < -box_half):   yy = yy + side
                        if (yy >  box_half):   yy = yy - side
                        if (zz < -box_half):   zz = zz + side
                        if (zz >  box_half):   zz = zz - side
                          
                        # distance between i and j
                        rd  = xx * xx + yy * yy + zz * zz
                        rij = sqrt(rd)
                           
                        bin = int(ceil(rij/dr)) # determine in which bin the distance falls
                        if (bin <= maxbin):
                            hist[bin] += 1

                # Reset Parameters
                atoms = []
                config += 1                                                                     # Update config counter

            continue                                                                            # We're not interested in non-COM atoms
        coords = map(float, (line[31:54]).split())                                              # However, if we recognize a COM site, map the cartesian coords
        atoms.append((coords[0], coords[1], coords[2]))

    traj.close()

    # Normalize
    print "Normalizing ... "
    nconfigs = config - 1
    avg_part = part_total / float(nconfigs)
    rho = avg_part/pow(side, 3.0)                                                               # Density
    norm = 2.0 * pi * dr * rho * nconfigs * avg_part

    print "NORM: %f\tDR: %f\tRHO: %f\t NCONFIGS: %d\tAVG_PART: %f" % (norm,dr,rho,nconfigs,avg_part)

    for i in range(1, maxbin+1):
        rrr = (i - 0.5) * dr
        val = hist[i]/ norm / ((rrr * rrr) + (dr * dr) / 12.0)
        rdf.update({rrr:val})
    
    return rdf

#-------------------------------------------------------------------#

# Write to file
pdbsdir = getcwd() # directory with PDB files
rdfout  = "./RDF.dat"

# Check Arguments
if len(sys.argv) == 4:
    trajfile      = sys.argv[1]
    boxlength     = float(sys.argv[2])
    n_bins        = int(sys.argv[3])
else:
    print "Insufficent number of arguments. USAGE: $ python RDF.py [traj_file.pdb] [box_length (A)] [# of bins]"
    sys.exit(1)

rdf = calcRDF(trajfile, pdbsdir, boxlength, n_bins)

print "Writing output file ... " + rdfout
rdfout = open(rdfout, "w")
for r in sorted(rdf.iterkeys()):                                                # Sort
    rdfout.write("%15.8g %15.8g\n"%(r, rdf[r]))
rdfout.close()



"""
    Models.Stochastic.MobileDispersion.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
from scipy.optimize import minimize

USE_SPLINE = True       # using spline while the curve model is not sufficient
M1_WEIGHT = 6
M2_WEIGHT = 4

class DispersionRatio:
    def __init__(self, rgmin, rgmax, N, T, poresize, t0, N0):
        self.poresize = poresize
        me = 1.5
        mp = 1.5
        rgs_ = np.linspace(rgmin, rgmax, 20)
        rhov = rgs_/poresize
        rhov[rhov > 1] = 1
        niv = N*(1 - rhov)**me
        tiv = T*(1 - rhov)**mp
        dsv = 2*niv*tiv**2
        dmv = (t0 + niv*tiv)**2/N0
        ratios = np.sqrt(dmv/dsv)
        if USE_SPLINE:
            self.spline = UnivariateSpline(rgs_, ratios, s=0)
        else:
            def objective(p):
                a, b = p
                y = a*np.power(self.poresize - rgs_, b)     # this curve model does not seem to be sufficient.
                return np.sum((y - ratios)**2)
            res = minimize(objective, [100,-2])
            self.params = res.x
            print("self.params=", self.params)
        self.rgs = rgs_
        self.ratios = ratios
        self.t0 = t0
        self.N0 = N0

    def __call__(self, rgs):
        if USE_SPLINE:
            return self.spline(rgs)
        else:
            a, b = self.params
            return a*np.power(self.poresize - rgs, b)

    def estimate_N0(self, mdscale):
        return self.N0/mdscale**2

    def estimete_tI(self, peak_rgs, props, target_moments_list, mdscale, N, K, x0, poresize, N0):
        me = 1.5
        mp = 1.5
        T = K/N
        rhov = peak_rgs/poresize
        rhov[rhov > 1] = 1

        def objective(p):
            tI = p[0]
            dev_list = []
            for k, (M, rho) in enumerate(zip(target_moments_list, rhov)):
                M1_ = x0 + N * T * (1 - rho)**(me + mp)
                M2_ = np.sqrt(2 * N * T**2 * (1 - rho)**(me + 2*mp) + (M1_ - tI)**2/N0)
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_- M[1])**2)
            return np.sum(np.asarray(dev_list)*props)
        
        init_tI = x0 - self.t0
        res = minimize(objective, [init_tI])
        tI = res.x[0]
        print("init_tI, tI=",init_tI, tI)
        return tI

    def covert_to_curveparams(self, peak_rgs, props, target_moments_list, colparams, scales, debug_info=None):
        from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCOL_PARAMS
        print("covert_to_curveparams: colparams=", colparams)
        N, K, x0, poresize, mdscale = colparams[0:NUM_SDMCOL_PARAMS]
        N0 = self.estimate_N0(mdscale)
        tI = self.estimete_tI(peak_rgs, props, target_moments_list, mdscale, N, K, x0, poresize, N0)
        curveparams = np.concatenate([[N, K, x0, poresize, N0, tI], scales])
        if debug_info is not None:
            from molass_legacy.Models.Stochastic.DispersiveUtils import compute_elution_curves
            import molass_legacy.KekLib.DebugPlot as dplt
            print("mdscale=", mdscale)
            x, y = debug_info
            cy_list, ty = compute_elution_curves(x, curveparams, peak_rgs)
            with dplt.Dp():
                fig, ax = dplt.subplots()
                ax.set_title("covert_to_curveparams")
                ax.plot(x, y, label='data')
                for k, cy in enumerate(cy_list):
                    ax.plot(x, cy, ':', label='rg=%g' % peak_rgs[k])
                ax.plot(x, ty, ':', label='model total', color='red')
                ax.legend()
                fig.tight_layout()
                ret = dplt.show()
                if not ret:
                    return
        return curveparams

def proof_demo():
    from molass_legacy.Models.Stochastic.DispersivePdf import N0, dispersive_monopore_pdf

    N = 2000
    T = 0.5
    K = N*T
    me = 1.5
    mp = 1.5
    poresize = 80
    rgs = np.flip(np.arange(10, poresize, 10))
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    t0 = 500
    tI = 0
    x = np.arange(0, 1400)

    dratio = DispersionRatio(rgs[-1], rgs[0], N, T, poresize, t0, N0)

    rgv = np.linspace(rgs[-1], rgs[0], 100)

    def compute_curves(x, N, T, x0, tI, N0, rhov):
        cy_list = []
        nt_list = []
        for rho in rhov:
            ni = N*(1 - rho)**me
            ti = T*(1 - rho)**mp
            nt_list.append((ni, ti))
            cy = dispersive_monopore_pdf(x - tI, ni, ti, N0, x0 - tI)
            cy_list.append(cy)
        return cy_list, nt_list

    fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(15,8))
    fig.suptitle("Analysis of Dispersions with Dispersion Ratio Spline in Stochastic Dispersive Model", fontsize=20)

    def plot_curves_and_dispersions(ax1, ax2, ax3, N0, plot_spline=True, ref_ax=None):
        ax1.set_title("Elution Curves when Poresize=%g, $N_0$=%d" % (poresize, N0), fontsize=14)
        ax1.axvline(x=t0, label="$t_0$", color='red', alpha=0.5)
        cy_list, nt_list = compute_curves(x, N, T, t0, tI, N0, rhov)
        for k, (cy, rg) in enumerate(zip(cy_list, rgs)):
            ax1.plot(x, cy, label='$R_g$=%g' % rg, alpha=0.5)
        ax1.legend(loc='upper left')
        if ref_ax is not None:
            ax1.set_ylim(ref_ax.get_ylim())

        ax2.set_title("Mobile / Stationary Dispersions vs. $R_g$'s", fontsize=14)
        ax2.set_xlabel("$R_g$")
        ax2.set_ylabel("Dispersion")
        ax2.invert_xaxis()

        dispersions = []
        for k, (ni, ti) in enumerate(nt_list):
            ds = 2*ni*ti**2
            dm = (t0 + ni*ti)**2/N0
            dispersions.append((ds, dm))
        dispersions = np.array(dispersions)
        sdisps = np.sqrt(dispersions[:,0])
        mdisps = np.sqrt(dispersions[:,1])
        ax2.plot(rgs, sdisps, "o", label="Stationary", alpha=0.5)
        ax2.plot(rgs, mdisps, "o", label="Mobile", alpha=0.5)
        real_ratios = mdisps/sdisps
        model_ratios = dratio(rgs)
        scale = np.average(real_ratios/model_ratios)
        ax2.plot(rgs, sdisps*dratio(rgs)*scale, "+", label="Mobile (estimated with the spline)", markersize=10)
        ax2.legend()

        if plot_spline:
            ax3.set_title("Mobile / Stationary Dispersion Ratio Spline", fontsize=14)
            ax3.set_xlabel("$R_g$")
            ax3.set_ylabel("Dispersion Ratio")
            ax3.invert_xaxis()

            ax3.plot(rgs, mdisps/sdisps, "o", color="C2")
            ax3.plot(rgv, dratio(rgv), ":", color="C2")
        else:
            ax3.set_axis_off()

    plot_curves_and_dispersions(*axes[0,:], N0)
    plot_curves_and_dispersions(*axes[1,:], N0//4, plot_spline=False, ref_ax=axes[0,0])

    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")
    import os
    import seaborn as sns
    sns.set_theme()

    proof_demo()


#-*- coding: utf-8 -*-
#!python
"""
    borrowed from
    https://stackoverflow.com/questions/20995045/how-to-include-image-in-message-box-using-ctypes-in-python/21051586#21051586
    and slightly modified
"""

from ctypes import *
from ctypes.wintypes import *
from os import path
import platform

#################################################################

RelPath = lambda file : path.join(path.dirname(path.abspath(__file__)), file)

#################################################################

GetModuleHandle = windll.kernel32.GetModuleHandleW
GetModuleHandle.restype = HMODULE
GetModuleHandle.argtypes = [LPCWSTR]

#################################################################

IMAGE_ICON = 1
LR_LOADFROMFILE = 0x00000010
LR_CREATEDIBSECTION = 0x00002000

LoadImage = windll.user32.LoadImageW
LoadImage.restype = HANDLE
LoadImage.argtypes = [HINSTANCE, LPCWSTR, UINT, c_int, c_int, UINT]

#################################################################

LRESULT = c_int64 if platform.architecture()[0] == "64bit" else c_long

SendMessage = windll.user32.SendMessageW
SendMessage.restype = LRESULT
SendMessage.argtypes = [HWND, UINT, WPARAM, LPARAM]

#################################################################

MB_OK = 0x00000000

MessageBox = windll.user32.MessageBoxW
MessageBox.restype  = c_int
MessageBox.argtypes = [HWND, LPCWSTR, LPCWSTR, UINT]

#################################################################

WH_CBT = 5
HCBT_ACTIVATE = 5
HOOKPROC = WINFUNCTYPE(LRESULT, c_int, WPARAM, LPARAM)

SetWindowsHookEx = windll.user32.SetWindowsHookExW
SetWindowsHookEx.restype = HHOOK
SetWindowsHookEx.argtypes = [c_int, HOOKPROC, HINSTANCE, DWORD]

#################################################################

CallNextHookEx = windll.user32.CallNextHookEx
CallNextHookEx.restype = LRESULT
CallNextHookEx.argtypes = [HHOOK, c_int, WPARAM, LPARAM]

#################################################################

GetCurrentThreadId = windll.kernel32.GetCurrentThreadId
GetCurrentThreadId.restype = DWORD
GetCurrentThreadId.argtypes = None

#################################################################

UnhookWindowsHookEx = windll.user32.UnhookWindowsHookEx
UnhookWindowsHookEx.restype = BOOL
UnhookWindowsHookEx.argtypes = [HHOOK]

#################################################################
# code starts here

def AlterIcon(_hWnd, lpszIcon):

  WM_SETICON = 0x0080
  ICON_BIG = 1

  hModel = GetModuleHandle(None)
  hIcon = LoadImage(hModel,
              RelPath(lpszIcon),
              IMAGE_ICON,
              0, 0,
              LR_LOADFROMFILE | LR_CREATEDIBSECTION)

  SendMessage(_hWnd, WM_SETICON, ICON_BIG, hIcon)

def AlterButton(_hWnd):
  #**********************************************************#
  # center button code
  WNDENUMPROC = WINFUNCTYPE(BOOL, HWND, LPARAM)

  EnumChildWindows = windll.user32.EnumChildWindows
  EnumChildWindows.restype = BOOL
  EnumChildWindows.argtypes = [HWND, WNDENUMPROC, LPARAM]

  GetClassName = windll.user32.GetClassNameW
  GetClassName.restype = HWND
  GetClassName.argtypes = [HWND, LPCWSTR, c_int]

  GetClientRect = windll.user32.GetClientRect 
  GetClientRect.restype = BOOL
  GetClientRect.argtypes = [HWND, POINTER(RECT)]

  MoveWindow = windll.user32.MoveWindow
  MoveWindow.restype = BOOL
  MoveWindow.argtypes = [HWND, c_int, c_int, c_int, c_int, BOOL]

  MapWindowPoints = windll.user32.MapWindowPoints
  MapWindowPoints.restype = c_int
  MapWindowPoints.argtypes = [HWND, HWND, POINTER(POINT), UINT]

  def EnumChildProc(hwnd, lParam):
    ClassName = (c_wchar * 7)()
    if GetClassName(hwnd, ClassName, 7) > 0:
      if ClassName.value.lower() == "button":
        wrect = RECT()
        GetClientRect(lParam, byref(wrect))
        brect = RECT()
        GetClientRect(hwnd, byref(brect))
        bpoint = RECT()
        MapWindowPoints(hwnd, lParam, cast(byref(bpoint), POINTER(POINT)), 2)
        MoveWindow(hwnd,
                  ((wrect.right - wrect.left) - (brect.right - brect.left)) // 2,
                  bpoint.top,
                  brect.right - brect.left,
                  brect.bottom - brect.top,
                  True)
        return False
    return True

  #**********************************************************#
  pEnumChildProc = WNDENUMPROC(EnumChildProc)
  EnumChildWindows(_hWnd, pEnumChildProc, _hWnd.value)
  #**********************************************************#

def MyMessageBox(hWnd, lpText, lpCaption, uType, hook_callback=None):
  hHook = HHOOK(None)
  #**********************************************************#

  def CBTProc(nCode, wParam, lParam):
    if nCode == HCBT_ACTIVATE:
      _hWnd = cast(wParam, HWND)
      if hook_callback is not None:
        hook_callback(_hWnd)
    CallNextHookEx(hHook, nCode, wParam, lParam)
    return 0

  # WARNING: don't pass HOOKPROC(CBTProc) directly to SetWindowsHookEx
  pCBTProc = HOOKPROC(CBTProc)

  hHook = SetWindowsHookEx(WH_CBT, pCBTProc, None, GetCurrentThreadId())
  ret = MessageBox(hWnd, lpText, lpCaption, uType)
  UnhookWindowsHookEx(hHook)
  return ret

# example of usage
if __name__ == '__main__':
    def hook_callback( hWnd ):
        AlterIcon(hWnd, "favicon.ico")
        AlterButton(hWnd)
    MyMessageBox(None, "Hello world!", "Title", MB_OK, hook_callback=hook_callback)


# coding: utf-8
"""

    Tutorials.LPM.py

    Copyright (c) 2020, SAXS Team, KEK-PF

"""
import numpy as np
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DataStructure.LPM import LPM_3d
from MatrixData import simple_plot_3d

"""
    borrowed at https://stackoverrun.com/ja/q/6260216
    See Also https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.patches.FancyArrowPatch.html
"""
class Arrow3D(FancyArrowPatch):
       def __init__(self, xs, ys, zs, *args, **kwargs):
           FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
           self._verts3d = xs, ys, zs

       def draw(self, renderer):
           xs3d, ys3d, zs3d = self._verts3d
           xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
           self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
           FancyArrowPatch.draw(self, renderer)

def demo_3d(xr):
    ecurve_y = xr.e_curve.y
    lpm = LPM_3d(xr.data, ecurve_y=ecurve_y, debug=True)

    q = xr.vector

    print('xr.data.shape=', xr.data.shape, 'q.shape=', q.shape)

    plt.push()
    fig = plt.figure(figsize=(48, 24))
    gs = GridSpec(2,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1], projection='3d')
    ax02 = fig.add_subplot(gs[0,2], projection='3d')
    ax11 = fig.add_subplot(gs[1,1], projection='3d')
    ax12 = fig.add_subplot(gs[1,2], projection='3d')
    axes = [ax00, ax01, ax02, ax11, ax12]

    scale, elev = 101, -3   # for 20200630_11
    # scale, elev = 51, -2.2    # for 20200630_12

    simple_plot_3d(ax00, xr.data, x=q, edgecolors='cyan', zorder=0)

    # kwargs = {'rcount':5, 'ccount':5}
    kwargs = {}
    corrected = lpm.data.copy()
    simple_plot_3d(ax01, corrected, x=q, edgecolors='cyan', zorder=0)

    BP = lpm.adjust_with_mf(xr.e_index, xr.e_curve)
    simple_plot_3d(ax02, lpm.data, x=q, edgecolors='cyan')

    bp_min = np.min(BP)
    bp_max = np.max(BP)
    print(bp_min, bp_max)

    w = -scale
    zmin = bp_min*(1 - w) + bp_max*w
    w = scale
    zmax = bp_min*(1 - w) + bp_max*w

    fig.tight_layout()
    fig.subplots_adjust(top=0.95, hspace=0.1)

    titles = ["Not corrected", "LPM corrected", "BPA corrected", "LPM corrected (Zoomed)", "BPA corrected  (Zoomed)"]

    for ax, title in zip(axes, titles):
        ax.set_title(title, fontsize=50)

    for ax in [ax11, ax12]:
        ax.set_zlim(zmin, zmax)

    simple_plot_3d(ax11, corrected, x=q, edgecolors='cyan', alpha=0.1, zorder=0, **kwargs)
    simple_plot_3d(ax12, lpm.data, x=q, edgecolors='cyan', alpha=0.1, zorder=0, **kwargs)

    for ax in axes[0:3]:
        ax.view_init(20, -30)
    for ax in axes[3:]:
        ax.view_init(elev, -30)

    ZP = np.zeros(BP.shape)

    simple_plot_3d(ax11, ZP, x=q, edgecolors='pink', color='red', alpha=0.3, zorder=1, **kwargs)
    simple_plot_3d(ax11, BP, x=q, edgecolors='green', color='green', alpha=0.3, zorder=1, **kwargs)
    simple_plot_3d(ax12, ZP, x=q, edgecolors='pink', color='red', alpha=0.3, zorder=1, **kwargs)

    je = len(ecurve_y)-1
    arrow_scale = 1
    for x_, y_, z_ in [ ([q[0],q[0]], [0,0], [BP[0,0]*arrow_scale, 0]),
                        ([q[0],q[0]], [je,je], [BP[0,-1]*arrow_scale, 0]),
                        ([q[-1],q[-1]], [0,0], [BP[-1,0]*arrow_scale, 0]),
                        ([q[-1],q[-1]], [je,je], [BP[-1,-1]*arrow_scale, 0]),
                        ]:
        arrow = Arrow3D(x_, y_, z_,
                        mutation_scale=10, lw=5, arrowstyle="-|>", color="r",
                        shrinkA=0, shrinkB=0, zorder=1)
        ax11.add_artist(arrow)

    plt.show()
    plt.pop()



"""

    Menus.JupyterMenu.py

    Copyright (c) 2022, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter import Tk

class JupyterMenu(Tk.Menu):
    def __init__(self, parent, menubar):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade(label="Jupyter", menu=self )
        self.add_command(label="Jupyter Lab", command=self.run_jupyterlab)

    def run_jupyterlab(self):
        from Jupyter.JupyterLab import run_jupyterlab
        self.entryconfig(0, state=Tk.DISABLED)
        self.update()
        run_jupyterlab(parent=self.parent)
        self.entryconfig(0, state=Tk.NORMAL)



"""
    obtained from https://github.com/GarmanGroup/CorMapAnalysis
    converted from Python2 to Python3
"""

"""CorMap class implementing analyis of results from the ATSAS suite program
DATCMP.
"""
import glob
import subprocess
import re
import os
import shutil
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd
import seaborn as sns


class ScatterAnalysis(object):
    """
    Scatter curve similarity analysis class.

    CorMap class implementing analysis of data output from a run of the
    ATSAS suite program DATCMP.
    """

    # ----------------------------------------------------------------------- #
    #                         CLASS VARIABLES                                 #
    # ----------------------------------------------------------------------- #
    PLOT_LABEL = {'family': 'serif',
                  'weight': 'normal',
                  'size': 16}
    mpl.rc('font', family='serif', weight='normal', size=12)

    PLOT_NUM = 0

    # ----------------------------------------------------------------------- #
    #                         CONSTRUCTOR METHOD                              #
    # ----------------------------------------------------------------------- #
    def __init__(self, intensities, scattering_angles, datcmp_data,
                 x_axis_vec=[], x_metric="", x_units=""):
        """
        Create a ScatterAnalysis object.

        Constructor method to create ScatterAnalysis object which exposes a
        simple API for analysing the DATCMP data.

        Parameters
        ----------
        intensities : 2D numpy array
            Array of intensity values where each column represents the
            intensity values for a single frame and each row represents the
            intensity at a particular scattering angle.
        scattering_angles : 1D numpy array
            Array of scattering angles measured for the SAXS run.
        datcmp_data : dictionary
            dictionary containing all of the data from the DATCMP output. The
            key type is a string of image pairs and the value type is a list of
            the pairwise correlation results from the analysis between the
            image pairs.
        x_axis_vec : numpy array, optional (default=[])
            Vector of values to use on the x-axis of plots. This can be things
            like dose values for example. If not specified then frame number is
            used as a default.
        x_metric : str, optional (default="")
            String specifying the type of metric used for the x-axis. An
            example would be "Dose". If not specified then the default metric
            used on the x-axis is "Frame Number"
        x_units : str, optional (default="")
            String specifying the units used for the x-axis metric. For example
            if the x-axis metric was the 'dose', then x_units may be specified
            as "kGy".
        """
        self.I = intensities
        self.q = scattering_angles
        self.datcmp_data = datcmp_data
        self.x_axis = x_axis_vec
        self.x_metric = x_metric
        self.x_units = x_units

    # ----------------------------------------------------------------------- #
    #                            CLASS METHODS                                #
    # ----------------------------------------------------------------------- #

    # Note that class methods can access class attributes but not instance
    # attributes.
    @classmethod
    def from_1d_curves(cls, scat_curve_location, x_axis_vec=[], metric="",
                       units="", first=-1, last=-1, smin=-1, smax=-1):
        """
        Create a ScatterAnalysis object from .

        Method to create ScatterAnalysis object from a set of 1D scattering
        curve files.

        Parameters
        ----------
        scattering_curve_location : str
            Location of the scattering curves for the dataset.
        x_axis : numpy array, optional (default=[])
            Vector of values to use on the x-axis of plots. This can be things
            like dose values for example. If not specified then frame number is
            used as a default.
        x_metric : str, optional (default="")
            String specifying the type of metric used for the x-axis. An
            example would be "Dose". If not specified then the default metric
            used on the x-axis is "Frame Number"
        x_units : str, optional (default="")
            String specifying the units used for the x-axis metric. For example
            if the x-axis metric was the 'dose', then x_units may be specified
            as "kGy".
        first : int, optional (default=-1)
            Index of the first point to be kept. This is mutually exclusive
            with argument parameter 'smin'. Default value is -1. This will
            ultimately result in the method not using this value.
        last : int, optional (default=-1)
            Index of the last point to be kept. This is mutually exclusive with
            argument parameter 'smax'. Default value is -1. This will
            ultimately result in the method not using this value.
        smin : float, optional (Default=-1)
            Minimal scattering angle value to be kept. This is mutually
            exclusive with argument parameter 'first'. Default value is -1.
            This will ultimately result in the method not using this value.
        smax : float, optional (Default=-1)
            Maximal scattering angle value to be kept. This is mutually
            exclusive with argument parameter 'last'. Default value is -1.
            This will ultimately result in the method not using this value.

        Returns
        -------
        ScatterAnalysis
            The ScatterAnalysis object

        Examples
        --------
        To create a ScatterAnalysis object that will plot only against frame
        numbers then the only input required is the location of the set of
        files:

        >>>  scat_obj = ScatterAnalysis.from_1d_curves("saxs_files.00*.dat")

        If you want to plot an x-axis that is different from the frame number,
        e.g. the dose, then we would specify something like:

        >>>  scat_obj = ScatterAnalysis.from_1d_curves("saxs_files.00*.dat", np.array([10, 20, 30, 40]), "Dose", "kGy")

        If you don't want to compare the entire range of the SAXS curve (this
        may be the case where there are sections of the 1D curves that are
        noisy) then there is the option to crop the curves before running the
        CorMap test to create the ScatterAnalysis object.
        In the example below we create a ScatterAnalysis object where we crop
        the first 99 frames and use data up to a scattering angle of 3 nm^{-1}.

        >>>  scat_obj = ScatterAnalysis.from_1d_curves("saxs_files.00*.dat", first=99, smax=3.0)

        Note
        ----
        The x_axis_vec array needs to be the same length as the total number of
        frames in the dataset otherwise the frame number will be used on the
        x-axis by default

        """
        # Go through files and extract the frame data.
        file_list = glob.glob(scat_curve_location)
        num_frames = len(file_list)

        # Decide whether the frames need to be cropped first.
        if (first < 0) and (last < 0) and (smin < 0) and (smax < 0):
            # Run DATCMP to get pairwise comparison information.
            datcmp_data = ScatterAnalysis.get_datcmp_info(scat_curve_location)
            scattering_angles = parse_saxs_dat_file(file_list[0], "scattering angle")
            intensities = np.zeros([len(scattering_angles), num_frames])
            num_data_points_per_curve = len(scattering_angles)
            print("Number of data points on resulting scattering profile: {}".format(num_data_points_per_curve))
            for i, file in enumerate(file_list):
                intensities[:, i] = parse_saxs_dat_file(file, "intensity", num_data_points_per_curve)
        else:
            if last > 0 and smax > 0:
                print("**************** WARNING ******************")
                print("Both 'last' and 'smax' are defined but they are mutually exclusive.")
                print("Only 'last' is being used!!!")
            if first > 0 and smin > 0:
                print("**************** WARNING ******************")
                print("Both 'first' and 'smin' are defined but they are mutually exclusive.")
                print("Only 'first' is being used!!!")
            temp_dir_name = "Temp_dir"
            temp_file_prefix = "temp"
            os.mkdir(temp_dir_name)
            crop_saxs_files(file_list, temp_dir_name, temp_file_prefix, first, last, smin, smax)
            # Run DATCMP to get pairwise comparison information.
            new_scat_curve_location = "{}/{}*".format(temp_dir_name,
                                                      temp_file_prefix)
            new_file_list = glob.glob(new_scat_curve_location)
            print("Number of files = {}".format(len(new_file_list)))
            datcmp_data = ScatterAnalysis.get_datcmp_info(new_scat_curve_location)
            scattering_angles = parse_saxs_dat_file(new_file_list[0],
                                                    "scattering angle")
            num_data_points_per_curve = len(scattering_angles)
            print("Number of data points on resulting scattering profile: {}".format(num_data_points_per_curve))
            intensities = np.zeros([num_data_points_per_curve,
                                    len(new_file_list)])
            for i, filename in enumerate(new_file_list):
                intensities[:, i] = parse_saxs_dat_file(filename, "intensity", num_data_points_per_curve)

            # Remove temporary directory if it still exists
            if os.path.exists(temp_dir_name):
                shutil.rmtree(temp_dir_name)

        # Organise the x-axis used for the plots. Default will be the frame
        # number.
        if not isinstance(x_axis_vec, list):
            if len(x_axis_vec) == num_frames:
                x_axis = x_axis_vec
            else:
                print("x_axis_vec is not the same length as the number of")
                print("frames. Using frame numbers instead.")
                x_axis = np.linspace(1, num_frames, num_frames)
        else:
            x_axis = np.linspace(1, num_frames, num_frames)

        if metric and units:
            x_metric = metric
            x_units = units
        else:
            x_metric = "Frame number"
            x_units = ""

        return cls(intensities, scattering_angles, datcmp_data, x_axis,
                   x_metric, x_units)

    @classmethod
    def get_datcmp_info(self, scattering_curve_files):
        """
        Extract the data produced by DATCMP.

        This method is used by the constructor method of the ScatterAnalysis
        class. It runs the DATCMP program on the dataset and returns a
        dictionary containing the main results from the DATCMP run.

        Parameters
        ----------
        scattering_curve_files : str
            Location of the scattering curves for the dataset.

        Returns
        -------
        dict(str, numpy.array)
            Dictionary containing the results of the DATCMP run. The dictionary
            key is a string (with no spaces) denoting the pair of frames that
            were compared e.g. "1,2" would be frames 1 and 2. The dictionary
            value is an array of DATCMP results for the corresponding pairwise
            comparison.

        Examples
        --------
        >>>  datcmp_data = scat_obj.get_datcmp_info("saxs_files.00*.dat")
        """
        cmd = "datcmp {}".format(scattering_curve_files)
        log = run_system_command(cmd)
        # define a dictionary to store the data produced from DATCMP - this
        # value will be overwritten.
        data_dict = {"1,2": 0}
        for line in iter(log.splitlines()):
            match_obj = re.match(r'\s* \d{1,} vs', line)
            if match_obj:
                data = line.split()
                if "*" in data[5]:
                    data[5] = data[5][:-1]
                data_dict["{},{}".format(data[0], data[2])] = [int(float(data[3])),
                                                               float(data[4]),
                                                               float(data[5])]
        return data_dict

    # ----------------------------------------------------------------------- #
    #                         INSTANCE METHODS                                #
    # ----------------------------------------------------------------------- #

    def find_diff_frames(self, frame=1, P_threshold=0.01, P_type="adjP"):
        """
        List all statistically dissimilar frames.

        This method finds all statistically dissimilar frames to any given
        frame using the CorMap test as outlined in Daniel Franke, Cy M Jeffries
        & Dmitri I Svergun (2015). The user can set the significance threshold
        as well as whether to use the Bonferroni corrected P values or not.
        (we recommend that you should use the Bonferroni corrected P values).

        Parameters
        ----------
        frame : int, optional (default=1)
            The frame that every other frame in the dataset is compared with.
            If not specified then all frames are compared to the first frame.
        P_threshold : float, optional (default=0.01)
            The significance threshold of the test. If it's not given then the
            default value is 1%.
        P_type : str, optional (default="adjP")
            String denoting whether to use the Bonferroni corrected P value
            (input string="adjP") or the ordinary P value (input string="P").
            Default is to use the Bonferroni corrected P value.

        Returns
        -------
        List
            A list of integers corresponding to all of the dissimilar frames

        Examples
        --------
        Find all frames that are dissimilar to frame 10

        >>>  diff_frames = scat_obj.find_diff_frames(frame=10)
        """
        if P_type == "adjP":
            p_col = 2
        elif P_type == "P":
            p_col = 1
        else:
            print("********************** ERROR ***************************")
            print("P_type '{}' Is not recognised".format(P_type))
            print("P_type can only take the values 'adjP' (default) or 'P'.")
        if frame <= self.I.shape[1]:
            diff_frames = []
            for i in range(0, self.I.shape[1]):
                if i+1 < frame:
                    key = "{},{}".format(i+1, frame)
                elif i+1 > frame:
                    key = "{},{}".format(frame, i+1)
                else:
                    continue
                significance_val = self.datcmp_data[key][p_col]
                if significance_val < P_threshold:
                    diff_frames.append(i+1)
            return diff_frames
        else:
            print("********************** ERROR ***************************")
            print("FRAME '{}' DOES NOT EXIST".format(frame))
            print("Use different frame numbers between 1 and {}".format(self.I.shape[1]))

    def find_first_n_diff_frames(self, n=1, frame=1, P_threshold=0.01,
                                 P_type="adjP"):
        """
        Find the first of n consecutive dissimilar frames.

        Return the first frame, F, where there are n-1 consecutive frames
        after F that are also statistically dissimilar from the chosen frame.

        Parameters
        ----------
        n : int, optional (default=1)
            The number of consecutive dissimilar frames to be considered
            significant.
        frame : int, optional (default=1)
            The frame that every other frame in the dataset is compared with.
            If not specified then all frames are compared to the first frame.
        P_threshold : float, optional (default=0.01)
            The significance threshold of the test. If it's not given then the
            default value is 1%.
        P_type : str, optional (default="adjP")
            string denoting whether to use the Bonferroni corrected P value
            (input string="adjP") or the ordinary P value (input string="P").
            Default is to use the Bonferroni corrected P value.

        Returns
        -------
        int
            Frame number of the first of n consecutive dissimilar frames.

        Examples
        --------
        Get the frame number that is the first of 3 consecutive dissimilar
        frames to frame 10.

        >>>  first_diff_frames = scat_obj.find_first_n_diff_frames(n=3, frame=10)
        """
        # Get list frames that are different
        list_of_diff_frames = self.find_diff_frames(frame, P_threshold, P_type)
        if n == 1:
            # If only looking for one frame then return the first value in
            # list of different frames.
            return list_of_diff_frames[0]
        elif n > 1:
            # If we're looking for more than one consecutive frame then we need
            # to keep track of the number of consecutive frames that we've
            # iterated through.
            consec_count = 0
            max_consec_count = 0
            fr_max_count = 0
            for i, curr_fr in enumerate(list_of_diff_frames):
                if i == 0:
                    prev_fr = curr_fr
                    consec_count = 1
                else:
                    if curr_fr == prev_fr + 1:
                        consec_count += 1
                    else:
                        consec_count = 1
                prev_fr = curr_fr
                if consec_count == n:
                    return curr_fr - n + 1
                if consec_count > max_consec_count:
                    max_consec_count = consec_count
                    fr_max_count = curr_fr - max_consec_count + 1
            print("************************ WARNING **************************")
            print("{} consecutive frames not reached!".format(n))
            print("The max number of consecutive frames was {}".format(max_consec_count))
            print("The initial frame for that run was frame {}.".format(fr_max_count))
        else:
            print("********************** ERROR ***************************")
            print("n MUST BE A POSITVE INTEGER VALUE")
            print("User chose n = {}.".format(n))
            print("Please choose a positve integer value for n.")

    def similar_frames(self, frame=1, P_threshold=0.01, P_type="adjP"):
        """
        List all statistically similar frames.

        This method finds all statistically similar frames to any given
        frame using the CorMap test as outlined in Daniel Franke, Cy M Jeffries
        & Dmitri I Svergun (2015). The user can set the significance threshold
        as well as whether to use the Bonferroni corrected P values or not.
        (we recommend that you should use the Bonferroni corrected P values).

        Parameters
        ----------
        frame : int, optional (default=1)
            The frame that every other frame in the dataset is compared with.
            If not specified then all frames are compared to the first frame.
        P_threshold : float, optional (default=0.01)
            The significance threshold of the test. If it's not given then the
            default value is 1%.
        P_type : str, optional (default="adjP")
            string denoting whether to use the Bonferroni corrected P value
            (input string="adjP") or the ordinary P value (input string="P").
            Default is to use the Bonferroni corrected P value.

        Returns
        -------
        List
            A list of integers corresponding to all of the similar frames

        Examples
        --------
        Find all frames that are similar to frame 10

        >>>  similar_frames = scat_obj.similar_frames(frame=10)
        """
        list_of_diff_frames = self.find_diff_frames(frame, P_threshold, P_type)
        return [i+1 for i in range(0, self.I.shape[1]) if i+1 not in list_of_diff_frames]

    def get_pw_data(self, frame1, frame2, datcmp_data_type="adj P(>C)"):
        """
        Get the CorMap results for a given pair of frames.

        Return C, P(>C) or Bonferroni adjusted P(>C) value from the DATCMP
        output.

        Parameters
        ----------
        frame1 : int
            number of the 1st frame used for the pairwise analyis
        frame2 : int
            number of the 2nd frame used for the pairwise analyis
        datcmp_data_type: str, optional (default="adj P(>C)")
            String specifying the pairwise result to be returned.
            The input options are:
            1) 'C' - This will return the C value i.e. the max observed patch
            of continuous runs of -1 or 1
            2) 'P(>C)' - This will return the P value of observing a patch of
            continuous runs of -1 or 1 bigger than the corresponding C value.
            3) 'adj P(>C)' - This will return the Bonferroni adjusted P value
            of observing a patch of continuous runs of -1 or 1 bigger than the
            corresponding C value.

        Returns
        -------
        int or float
            A value specifying the largest number of consecutive positive or
            negative values in a row, C, or the probability value (either
            standard P or Bonferroni corrected) the we observed more than C
            positive or negative values in a row.

        Examples
        --------
        Get the number of the longest run of positive or negative values in a
        row between frames 1 and 2

        >>>  C = scat_obj.get_pw_data(1, 2, datcmp_data_type='C')

        Get the Bonferroni corrected probability that we would observed a run
        of positive or negative values greater than C

        >>>  C = scat_obj.get_pw_data(1, 2, datcmp_data_type='adj P(>C)')
        """
        if datcmp_data_type == "C" or datcmp_data_type == 0:
            dat_type = 0
        elif datcmp_data_type == "P(>C)":
            dat_type = 1
        elif datcmp_data_type == "adj P(>C)":
            dat_type = 2
        else:
            print("********************** ERROR ***************************")
            print("INVALID DATCMP DATA TYPE CHOSEN: '{}' DOES NOT EXIST".format(datcmp_data_type))
            print("Please choose either 'C', 'P(>C)' or 'adj P(>C)'.")

        if frame1 < frame2:
            datcmp_key = "{},{}".format(frame1, frame2)
        elif frame2 < frame1:
            datcmp_key = "{},{}".format(frame2, frame1)

        if datcmp_key in self.datcmp_data:
            return self.datcmp_data[datcmp_key][dat_type]
        else:
            print("********************** ERROR ***************************")
            print("KEY '{}' DOES NOT EXIST".format(datcmp_key))
            print("Use different frame numbers between 1 and {}".format(self.I.shape[1]))

    def calc_cormap(self):
        """
        Calculate correlation map.

        Method to calculate the full correlation map for entire dataset.

        Parameters
        ----------
        N/A

        Returns
        -------
        numpy 2D array
            2D matrix of correlations for all SAXS frames

        Examples
        --------
        >>>  correlation_matrix = scat_obj.calc_cormap()
        """
        return np.corrcoef(self.I)

    def calc_pwcormap(self, frame1, frame2):
        """
        Calculate pairwise correlation map between two frames.

        Calculation of the pairwise correlation map between two chosen frames.

        Parameters
        ----------
        frame1 : int
            Number of the 1st frame used for the pairwise analyis
        frame2 : int
            Number of the 2nd frame used for the pairwise analyis

        Returns
        -------
        2D Numpy array
            Array with the correlation between two frames.

        Examples
        --------
        Get the correlation map between frames 1 and 10.

        >>>  pairwise_corr_map = scat_obj.calc_pwcormap(1, 10)
        """
        pw_I = np.column_stack([self.I[:, frame1-1], self.I[:, frame2-1]])
        return np.corrcoef(pw_I)

    def get_pw_data_array(self, frame=0, delete_zero_row=True):
        """
        Get C, P(>C) and Bonferroni corrected P(>C) values for a given frame.

        Return an array of all C, P(>C) and Bonferroni adjusted P(>C) values
        from the DATCMP output for the requested frame. E.g. if you choose
        frame 1 then this means that the method with return an array of C,
        P(>C) and Bonferroni adjusted P(>C) values calculated between frame 1
        and all other frames.

        Parameters
        ----------
        frame : int, optional (default=0)
            The frame that every other frame in the dataset is compared with.
            If not specified then ALL data from the DATCMP results are
            returned.
        delete_zero_row : bool, optional (default=True)
            Choose whether to include a row where the chosen frame would've
            been compared with itself - although this row is filled with zero.
            This makes sure that you'll return an array whose number of rows
            is the same as the number of frames. By default this is set to
            false.

        Returns
        -------
        2D Numpy array
            Array of all C, P(>C) and Bonferroni adjusted P(>C) values from the
            DATCMP results for the requested frame.

        Examples
        --------
        Get all DATCMP data for when all other frames are compared with frame
        10

        >>>  datcmp_data = scat_obj.get_pw_data_array(frame=10)
        """
        if frame == 0:
            pw_data = np.zeros([len(self.datcmp_data), 3])
            for i, values in enumerate(self.datcmp_data.values()):
                pw_data[i, :] = np.asarray(values)
        elif 1 <= frame <= self.I.shape[1]:
            pw_data = np.zeros([self.I.shape[1], 3])
            for i in range(0, self.I.shape[1]):
                if i+1 < frame:
                    key = "{},{}".format(i+1, frame)
                elif i+1 > frame:
                    key = "{},{}".format(frame, i+1)
                else:
                    continue
                pw_data[i, :] = np.asarray(self.datcmp_data[key])
        else:
            print("********************** ERROR ***************************")
            print("FRAME '{}' DOES NOT EXIST".format(frame))
            print("Use a frame number between 1 and {}".format(self.I.shape[1]))

        if delete_zero_row and frame > 0:
            return np.delete(pw_data, (frame-1), axis=0)
        else:
            return pw_data


# ----------------------------------------------------------------------- #
#                        PLOT THE CORRELATION MAP                         #
# ----------------------------------------------------------------------- #
    def plot_cormap(self, colour_scheme="gray", display=True, save=False,
                    filename="", directory=""):
        """
        Plot the full correlation map.

        Plot the correlation map of all frames in the dataset.

        Parameters
        ----------
        colour_scheme : str, optional (default="gray")
            The colour scheme used to plot the correlation map. The list of
            schemes can be found on the relevant matplotlib webpage:
            http://matplotlib.org/examples/color/colormaps_reference.html
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.

        Examples
        --------
        Plot the full correlation map

        >>>  scat_obj.plot_cormap()

        Save the correlation map in the current working directory without
        displaying it

        >>>  scat_obj.plot_cormap(display=False, save=True, filename="MyPlot.png")
        """
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        min_q = min(self.q)
        max_q = max(self.q)
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          PLOT CORMAP                            #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        plt.figure(self.PLOT_NUM)
        plt.gca().xaxis.grid(False)
        plt.gca().yaxis.grid(False)
        cormap = plt.imshow(self.calc_cormap(), cmap=colour_scheme,
                            extent=[min_q, max_q, min_q, max_q])
        plt.xlabel(r'Scattering Vector, q (nm$^{-1}$)',
                   fontdict=self.PLOT_LABEL)
        plt.ylabel(r'Scattering Vector, q (nm$^{-1}$)',
                   fontdict=self.PLOT_LABEL)
        plt.colorbar(cormap)

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

# ----------------------------------------------------------------------- #
#                    PLOT THE PAIRWISE CORRELATION MAP                    #
# ----------------------------------------------------------------------- #
    def plot_pwcormap(self, fr1, fr2, colour_scheme="gray", display=True,
                      save=False, filename="", directory=""):
        """
        Plot the pairwise correlation map between two chosen frames.

        Plot the pairwise correlation map between two frames fr1 and fr2.

        Parameters
        ----------
        fr1 : int
            Number of the 1st frame used for the pairwise analyis
        fr2 : int
            Number of the 2nd frame used for the pairwise analyis
        colour_scheme : str, optional (default="gray")
            The colour scheme used to plot the correlation map. The list of
            schemes can be found on the relevant matplotlib webpage:
            http://matplotlib.org/examples/color/colormaps_reference.html
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.

        Examples
        --------
        Plot the pairwise correlation map between frame 1 and frame 2

        >>>  scat_obj.plot_pwcormap(1, 2)

        Save the pairwise correlation map in the current working directory
        without displaying it

        >>>  scat_obj.plot_pwcormap(1, 2, display=False, save=True, filename="MyPlot.png")
        """
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        min_q = min(self.q)
        max_q = max(self.q)
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                     PLOT PAIRWISE CORMAP                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        fig = plt.figure(self.PLOT_NUM)
        plt.gca().xaxis.grid(False)
        plt.gca().yaxis.grid(False)
        cormap = plt.imshow(self.calc_pwcormap(frame1=fr1, frame2=fr2),
                            cmap=colour_scheme,
                            extent=[min_q, max_q, min_q, max_q])
        plt.xlabel(r'Scattering Vector, q (nm$^{-1}$)',
                   fontdict=self.PLOT_LABEL)
        plt.ylabel(r'Scattering Vector, q (nm$^{-1}$)',
                   fontdict=self.PLOT_LABEL)
        plt.colorbar(cormap, ticks=[1, -1])
        adjP = self.get_pw_data(fr1, fr2, "adj P(>C)")
        C = self.get_pw_data(fr1, fr2, "C")
        if self.x_units:
            change_in_x = abs(self.x_axis[fr1-1] - self.x_axis[fr2-1])
            plt.title(r'PW CorMap: frame {} vs {}.C = {}, adj P(>C) = {}, $\Delta${} = {:.2f} {}'.format(fr1, fr2, C, adjP, self.x_metric, change_in_x, self.x_units), ha='center')
        else:
            plt.title(r'Pairwise CorMap: frame {} vs {}. C = {}, adj P(>C) = {}'.format(fr1, fr2, C, adjP), ha='center')

        fig.canvas.mpl_connect('draw_event', on_draw)
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

    # ----------------------------------------------------------------------- #
    #                  PLOT THE HISTOGRAM OF PAIRWISE DATA                    #
    # ----------------------------------------------------------------------- #
    def plot_histogram(self, frame=0, datcmp_data_type="C", display=True,
                       save=False, filename="", directory="", num_bins=20):
        """
        Plot the histogram of the DATCMP data.

        For any chosen frame this methods plots a histogram of the DATCMP data
        for all pairwise comparisons with that frame. If no frames are chosen
        then the method plots all pairwise frame comparison data.

        Parameters
        ----------
        frame : int, optional (default=0)
            Number of the frame for which all pairwise comparison data is
            plotted. If not specified then all pairwise frame comparison data
            is shown.
        datcmp_data_type : str, optional (default="C")
            The data type from the DATCMP results that is plotted.
            The input options are:
            1) 'C' - This will return the C value i.e. the max observed patch
            of continuous runs of -1 or 1
            2) 'P(>C)' - This will return the P value of observing a patch of
            continuous runs of -1 or 1 bigger than the corresponding C value.
            3) 'adj P(>C)' - This will return the Bonferroni adjusted P value
            of observing a patch of continuous runs of -1 or 1 bigger than the
            corresponding C value.
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.
        num_bins : int, optional (default=20)
            Number of bins used to plot in the histogram.

        Examples
        --------
        Plot histogram of all of the DATCMP data with data type "C" (as defined
        above)

        >>>  scat_obj.plot_histogram()

        Plot histogram of all of Bonferroni corrected P values with all frames
        compared to frame 10.

        >>> scat_obj.plot_histogram(frame=10, datcmp_data_type="adj P(>C)")

        Save the histogram in the current working directory without displaying
        it

        >>>  scat_obj.plot_histogram(display=False, save=True, filename="MyPlot.png")
        """
        if datcmp_data_type == "C" or datcmp_data_type == 0:
            dat_type = 0
        elif datcmp_data_type == "P(>C)":
            dat_type = 1
        elif datcmp_data_type == "adj P(>C)":
            dat_type = 2
        else:
            print("********************** ERROR ***************************")
            print("INVALID DATCMP DATA TYPE CHOSEN: '{}' DOES NOT EXIST".format(datcmp_data_type))
            print("Please choose either 'C', 'P(>C)' or 'adj P(>C)'.")
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          PLOT HISTOGRAM                         #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        plt.figure(self.PLOT_NUM)
        plt.hist(self.get_pw_data_array(frame)[:, dat_type], bins=num_bins)
        plt.xlabel("{}".format(datcmp_data_type), fontdict=self.PLOT_LABEL)
        plt.ylabel(r'Frequency', fontdict=self.PLOT_LABEL)
        if frame == 0:
            plt.title("{} values for all pairwise comparisons".format(datcmp_data_type))
        else:
            plt.title("{} values for all pairwise comparisons with frame {}".format(datcmp_data_type, frame))

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

    # ----------------------------------------------------------------------- #
    #                    SCATTER PLOT OF PAIRWISE DATA                        #
    # ----------------------------------------------------------------------- #
    def plot_scatter(self, frame=1, P_threshold=0.01, markersize=60,
                     display=True, save=False, filename="", directory="",
                     legend_loc="upper left", x_change=False, use_adjP=True,
                     xaxis_frame_num=False,
                     colours=["#0072B2", "#009E73", "#D55E00"],
                     markers=["o", "o", "o"]):
        """
        Create scatter plot of the frame similarity data.

        For any chosen frame this methods creates a scatter plot of the C
        values - the maximum observed patch of continuous runs of -1 or 1 - for
        a chosen frame against all other frames. The probabilities of those C
        values occuring by chance are also coloured accordingly, and this can
        be changed by the user.

        Parameters
        ----------
        frame : int, optional (default=0)
            Number of the frame for which all pairwise comparison data is
            plotted. If not specified then all pairwise frame comparison data
            against frame 1 is plotted.
        P_threshold : float, optional (default=0.01)
            The significance level for the data. If the probability is below
            this value then the current frame is considered dissimilar to the
            frame to which it is being compared. Default significance level is
            0.01.
        markersize : float, optional (default=60)
            Change the size of the markers on the plot. Default value is 60.
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.
        legend_loc : str or int, optional (default="upper left")
            Set the location of the legend in the figure. All possible values
            can be found in the Matplotlib Legend API documentation:
            http://matplotlib.org/api/legend_api.html. Default is set to the
            upper left of the figure.
        x_change : bool, optional (default=False)
            This option changes whether the x-axis is plotted with absolute
            values (x_change=True) or if it's plotted with differences between
            the x-axis values (x_change=False). For example if we compare
            choose the reference frame to be frame 10 (frame=10) then compare
            this with frame 30 then setting x_change=True will put the relevant
            point at frame 30. However if we set x_change=False this will put
            the relevant point to be at abs(10 - 30) = 20. This option allows
            you to determine if the dissimilar frames are frames that are
            collected far from the current frame. Default is set to
            x_change=True.
        use_adjP : bool, optional (default=True)
            Choose whether to use the Bonferroni corrected P values or to use
            the unadjusted P values. Default is to use the Bonferroni corrected
            P values.
        xaxis_frame_num : bool, optional (default=True)
            Choose whether to plot the x-axis with the frame number or whether
            to use the custom x-axis metric values that may (or may not) have
            been defined when the ScatterAnalysis object was constructed.
            Default is to use the custom metric (if it was defined).
        colours : list, optional (default=["#0072B2", "#009E73", "#D55E00"])
            Choose the colour scheme for the markers in the scatter plot, which
            is differentiated by the P value.
        markers : list, optional (default=["o", "o", "o"])
            Choose the markers used for the markers in the scatter plot, which
            is differentiated by the P value. A list of the markers that can be
            used can be found on the Matplotlib Markers API webpage:
            http://matplotlib.org/api/markers_api.html


        Examples
        --------
        Create the scatter plot of the data compared against frame 1

        >>>  scat_obj.plot_scatter()

        Create scatter plot where all frames are compared against frame 30 and
        the x axis plots the difference in frame number from frame 30.

        >>> scat_obj.plot_scatter(frame=30, x_change=True)

        Save the scatter plot in the current working directory without
        displaying it.

        >>>  scat_obj.plot_scatter(display=False, save=True, filename="MyPlot.png")
        """
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SCATTER PLOT                           #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        pwframe_data = self.get_pw_data_array(frame=frame,
                                              delete_zero_row=False)

        if xaxis_frame_num:
            x_axis = np.linspace(1, self.I.shape[1], self.I.shape[1])
        else:
            x_axis = self.x_axis

        if x_change:
            sub = x_axis[frame - 1]
            pwframe_data = np.column_stack([abs(x_axis - sub), pwframe_data])
        else:
            pwframe_data = np.column_stack([x_axis, pwframe_data])
        pwframe_data = np.delete(pwframe_data, (frame-1), axis=0)

        if use_adjP:
            lb_dict = {0: [colours[0], "P(>C) = 1"],
                       1: [colours[1], "{} <= P(>C) < 1".format(P_threshold)],
                       2: [colours[2], "P(>C) < {}".format(P_threshold)]}
            P_col = 3
        else:
            lb_dict = {0: [colours[0], "P(>C) == 1"],
                       1: [colours[1], "{} <= P(>C) < 1".format(P_threshold)],
                       2: [colours[2], "P(>C) < {}".format(P_threshold)]}
            P_col = 2

        plt.figure(self.PLOT_NUM)
        good_points = pwframe_data[pwframe_data[:, P_col] == 1]
        plt.scatter(good_points[:, 0], good_points[:, 1], color=lb_dict[0][0],
                    s=markersize, edgecolors='#ffffff', alpha=1,
                    marker=markers[0], label=lb_dict[0][1])

        ok_points = pwframe_data[1 > pwframe_data[:, P_col]]
        ok_points = ok_points[ok_points[:, P_col] >= P_threshold]
        plt.scatter(ok_points[:, 0], ok_points[:, 1], color=lb_dict[1][0],
                    s=markersize, edgecolors='#ffffff', alpha=1,
                    marker=markers[1], label=lb_dict[1][1])

        bad_points = pwframe_data[pwframe_data[:, P_col] < P_threshold]
        plt.scatter(bad_points[:, 0], bad_points[:, 1], color=lb_dict[2][0],
                    s=markersize, edgecolors='#ffffff', alpha=1,
                    marker=markers[2], label=lb_dict[2][1])

        plt.legend(loc=legend_loc, scatterpoints=1)
        if x_change:
            if xaxis_frame_num:
                plt.xlabel(r'$\Delta${}'.format("Frame Number"),
                               fontdict=self.PLOT_LABEL)
            elif self.x_units:
                plt.xlabel(r'$\Delta${} ({})'.format(self.x_metric, self.x_units),
                               fontdict=self.PLOT_LABEL)
            else:
                plt.xlabel(r'$\Delta${}'.format(self.x_metric),
                               fontdict=self.PLOT_LABEL)
        else:
            if xaxis_frame_num:
                plt.xlabel("Frame Number", fontdict=self.PLOT_LABEL)
            elif self.x_units:
                plt.xlabel("{} ({})".format(self.x_metric, self.x_units),
                               fontdict=self.PLOT_LABEL)
            else:
                plt.xlabel("{}".format(self.x_metric),
                               fontdict=self.PLOT_LABEL)
        plt.ylabel(r'C', fontdict=self.PLOT_LABEL)
        # if xaxis_frame_num:
        #     ax1.set_title("C values against frame number for frame {}".format(frame))
        # else:
        #     ax1.set_title("C values against {} for frame {}".format(self.x_metric,
        #                                                         frame))

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

    # ----------------------------------------------------------------------- #
    #                           P(>C) HEAT MAP                                #
    # ----------------------------------------------------------------------- #
    def plot_heatmap(self, P_threshold=0.01, markersize=60,
                     display=True, save=False, filename="", directory="",
                     legend_loc=2, x_change=False, use_adjP=True,
                     xaxis_frame_num=False, P_values=True,
                     colours=["#0072B2", "#009E73", "#D55E00"]):
        """
        Create Heatmap of the frame similarity data.

        The heatmap provides a visual representation of the pairwise similarity
        analyis of the SAXS dataset.

        Parameters
        ----------
        P_threshold : float, optional (default=0.01)
            The significance level for the data. If the probability is below
            this value then one frame is considered dissimilar to the
            frame to which it is being compared. Default significance level is
            0.01.
        markersize : float, optional (default=60)
            Change the size of the markers on the plot. Default value is 60.
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.
        legend_loc : str or int, optional (default=2)
            Set the location of the legend in the figure. All possible values
            can be found in the Matplotlib Legend API documentation:
            http://matplotlib.org/api/legend_api.html. Default is set to the
            upper left (integer value is 2) of the figure.
        x_change : bool, optional (default=False)
            This option changes whether the x-axis is plotted with absolute
            values (x_change=True) or if it's plotted with differences between
            the x-axis values (x_change=False). For example if we compare
            choose the reference frame to be frame 10 (frame=10) then compare
            this with frame 30 then setting x_change=True will put the relevant
            point at frame 30. However if we set x_change=False this will put
            the relevant point to be at abs(10 - 30) = 20. This option allows
            you to determine if the dissimilar frames are frames that are
            collected far from the current frame. Default is set to
            x_change=True.
        use_adjP : bool, optional (default=True)
            Choose whether to use the Bonferroni corrected P values or to use
            the unadjusted P values. Default is to use the Bonferroni corrected
            P values.
        xaxis_frame_num : bool, optional (default=True)
            Choose whether to plot the x-axis with the frame number or whether
            to use the custom x-axis metric values that may (or may not) have
            been defined when the ScatterAnalysis object was constructed.
            Default is to use the custom metric (if it was defined).
        P_values : bool, optional (default=True)
            Choose whether to plot probability values or C values - the maximum
            observed patch of continuous runs of -1 or 1. Default is to use the
            P values.
        colours : list, optional (default=["#0072B2", "#009E73", "#D55E00"])
            Choose the colour scheme for the markers in the scatter plot, which
            is differentiated by the P value.

        Examples
        --------
        Create heatmap.

        >>>  scat_obj.plot_heatmap()

        Save the heatmap in the current working directory withot displaying it.

        >>>  scat_obj.plot_heatmap(display=False, save=True, filename="MyPlot.png")
        """
        full_data = []
        num_frames = self.I.shape[1]
        for frame in range(1, num_frames+1):
            pwframe_data = self.get_pw_data_array(frame=frame,
                                                  delete_zero_row=False)

            if xaxis_frame_num:
                x_axis = np.linspace(1, self.I.shape[1], self.I.shape[1])
            else:
                x_axis = self.x_axis

            if x_change:
                sub = x_axis[frame - 1]
                pwframe_data = np.column_stack([abs(x_axis - sub), pwframe_data])
            else:
                pwframe_data = np.column_stack([x_axis, pwframe_data])
            pwframe_data = np.delete(pwframe_data, (frame-1), axis=0)
            if use_adjP:
                P_col = 3
            else:
                P_col = 2

            good_points = pwframe_data[pwframe_data[:, P_col] == 1]
            ok_points = pwframe_data[1 > pwframe_data[:, P_col]]
            ok_points = ok_points[ok_points[:, P_col] >= P_threshold]
            bad_points = pwframe_data[pwframe_data[:, P_col] < P_threshold]

            xOrder = list(good_points[:, 0]) + list(ok_points[:, 0]) + list(bad_points[:, 0])
            C_values = list(good_points[:, 1]) + list(ok_points[:, 1]) + list(bad_points[:, 1])
            xData = [-1]*len(good_points[:, 0]) + [0]*len(ok_points[:, 0]) + [1]*len(bad_points[:, 0])

            if P_values:
                xOrder_sorted, xData_sorted = (list(t) for t in zip(*sorted(zip(xOrder, xData))))
                full_data.append(xData_sorted)

            else:
                xOrder_sorted, C_values_sorted = (list(t) for t in zip(*sorted(zip(xOrder, C_values))))
                full_data.append(C_values_sorted)

        full_DataFrame = pd.DataFrame(data=full_data,
                                      columns=xOrder_sorted,
                                      index=list(range(1, num_frames+1)))
        heatmap = plt.figure()
        ax = plt.subplot(111)

        if P_values:
            sns.heatmap(full_DataFrame, cmap=mpl.colors.ListedColormap(colours), cbar=False)

            # create legend information
            if use_adjP:
                good_label = "adj P(>C) == 1"
                ok_label = "{} <= adj P(>C) < 1".format(P_threshold)
                bad_label = "adj P(>C) < {}".format(P_threshold)
                plot_title = "adj P(>C) values for varying frame number"
            else:
                good_label = "P(>C) == 1"
                ok_label = "{} <= P(>C) < 1".format(P_threshold)
                bad_label = "P(>C) < {}".format(P_threshold)
                plot_title = "P(>C) values for varying frame number"
            good_patch = mpl.patches.Patch(color=colours[0], label=good_label)
            ok_patch = mpl.patches.Patch(color=colours[1], label=ok_label)
            bad_patch = mpl.patches.Patch(color=colours[2], label=bad_label)
            box = ax.get_position()
            ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
            lgd = ax.legend(handles=[good_patch, ok_patch, bad_patch], bbox_to_anchor=(1, 1), loc=legend_loc)

        else:
            sns.heatmap(full_DataFrame, cmap="YlGnBu", cbar=True)
            plot_title = "C values for varying frame number"

        if xaxis_frame_num:
            plt.xlabel('Frame Number', fontdict=self.PLOT_LABEL)
        else:
            plt.xlabel('Frame Number', fontdict=self.PLOT_LABEL)
            # if x_change:
            #     if self.x_units:
            #         plt.xlabel(r'$\Delta${} ({})'.format(self.x_metric, self.x_units),
            #                    fontdict=self.PLOT_LABEL)
            #     else:
            #         plt.xlabel(r'$\Delta${}'.format(self.x_metric),
            #                    fontdict=self.PLOT_LABEL)
            # else:
            #     if self.x_units:
            #         plt.xlabel("{} ({})".format(self.x_metric, self.x_units),
            #                    fontdict=self.PLOT_LABEL)
            #     else:
            #         plt.xlabel("{}".format(self.x_metric),
            #                    fontdict=self.PLOT_LABEL)
        plt.ylabel('Frame Number', fontdict=self.PLOT_LABEL)
        plt.title(plot_title)
        tick_labels = np.linspace(1, num_frames, 10).astype(int)
        plt.yticks(tick_labels[::-1], tick_labels)
        plt.xticks(tick_labels, tick_labels, rotation='horizontal')

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path, bbox_extra_artists=(lgd,),
                        bbox_inches='tight')
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            heatmap.show()

    # ----------------------------------------------------------------------- #
    #                        PLOT 1D INTENSITY CURVE                          #
    # ----------------------------------------------------------------------- #
    def plot_1d_intensity(self, frames, start_point=1, end_point=-1,
                          log_intensity=True, display=True, save=False,
                          filename="", directory="", legend_loc="upper right",
                          markersize=8):
        """
        Plot 1D scatter curves.

        Choose frames for which to plot the 1D scatter curves in the same
        figure.

        Parameters
        ----------
        frames : int or list
            Number of the frame or list of frame numbers to be plotted on the
            figure.
        start_point : int, optional (default=1)
            The point on the scatter curve from which you would like to start
            the plot. Default is to plot from the first point.
        end_point : int, optional (default=-1)
            The point on the scatter curve from which you would like to end
            the plot. Default is to plot up until to very last point on the
            curve.
        log_intensity : bool, optional (default=True)
            Choose whether to plot the intensity curve on a logarithmic scale
            (default) or on an absolute scale.
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.
        legend_loc : str or int, optional (default="upper left")
            Set the location of the legend in the figure. All possible values
            can be found in the Matplotlib Legend API documentation:
            http://matplotlib.org/api/legend_api.html. Default is set to the
            upper left of the figure.
        markersize : float, optional (default=60)
            Change the size of the markers on the plot. Default value is 60.


        Examples
        --------
        Plot the 1D intensity curve for frames 1, 10 and 100

        >>>  scat_obj.plot_1d_intensity([1, 10, 100], start_point=200, end_point=400)

        Plot the 1D intensity curve for frames 1, 10 and 100 between point 200
        and 400.

        >>>  scat_obj.plot_1d_intensity([1, 10, 100])

        Save the 1D intensity plot in the current working directory without
        displaying it.

        >>>  scat_obj.plot_1d_intensity(display=False, save=True, filename="MyPlot.png")
        """
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                        PLOT INTENSITY CURVE                     #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if end_point == -1:
            end_point = len(self.q)
        reciprocal_resolution = self.q[start_point-1:end_point]
        if isinstance(frames, list):
            frames = [i - 1 for i in frames]
        intensity = self.I[start_point-1:end_point, frames]
        if log_intensity:
            intensity = np.log(intensity)
        plt.figure(self.PLOT_NUM)
        if len(intensity.shape) == 2:
            for i in range(0, intensity.shape[1]):
                if self.x_units:
                    plt.plot(reciprocal_resolution, intensity[:, i], 'o',
                             markersize=markersize, markeredgecolor='#ffffff',
                             markeredgewidth=0.2,
                             label="Frame {}, {}={:.2f} {}".format(frames[i] + 1,
                                                                   self.x_metric,
                                                                   self.x_axis[frames[i]],
                                                                   self.x_units))
                else:
                    plt.plot(reciprocal_resolution, intensity[:, i], 'o',
                             label="Frame {}".format(frames[i] + 1))
        else:
            if self.x_units:
                plt.plot(reciprocal_resolution, intensity, 'o',
                         label="Frame {}, {}={:.2f} {}".format(frames,
                                                               self.x_metric,
                                                               self.x_axis[frames-1],
                                                               self.x_units))
            else:
                plt.plot(reciprocal_resolution, intensity, 'o',
                         label="Frame {}".format(frames))
        plt.xlabel(r'Scattering Vector, q ($nm^{-1}$)',
                   fontdict=self.PLOT_LABEL)
        if log_intensity:
            plt.ylabel('log(I) (arb. units.)', fontdict=self.PLOT_LABEL)
        else:
            plt.ylabel('Intensity (arb. units.)', fontdict=self.PLOT_LABEL)
        plt.title('1D Scattering Curve', fontdict=self.PLOT_LABEL)
        plt.legend(loc=legend_loc, scatterpoints=1)

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

    # ----------------------------------------------------------------------- #
    #              PLOT FIRST N DIFFERENT FRAMES FOR EACH FRAME               #
    # ----------------------------------------------------------------------- #
    def plot_first_n_diff_frames(self, n=1, P_threshold=0.01,
                                 P_type="adjP", display=True, save=False,
                                 filename="", directory=""):
        """
        Plot dissimilar frames

        For each frame in the dataset you can perform a similarity analysis
        against every other frame. From this you can calculate the first frame
        for which the probability value (P(>C) or adj P(>C)) for n consecutive
        frames is below the significance value (signifying that the frames are
        likely to be dissimilar). Note here that n and the P_threshold
        (significance value) are user defined values.
        This methods plots the first frame in that consecutive run for each
        'reference' frame in the dataset. By reference frame, we mean the frame
        for which every other frame is compared.

        Parameters
        ----------
        n : int, optional (default=1)
            Number of consecutive dissimilar frames to be considered
            dissimilar.
        P_threshold : float, optional (default=0.01)
            The significance level for the data. If the probability is below
            this value then one frame is considered dissimilar to the
            frame to which it is being compared. Default significance level is
            0.01.
        P_type : str, optional (default="adjP")
            String denoting whether to use the Bonferroni corrected P value
            (input string="adjP") or the ordinary P value (input string="P").
            Default is to use the Bonferroni corrected P value.
        display : bool, optional (default=True)
            Choose whether to display the plot. By default it is displayed.
        save : bool, optional (default=True)
            Choose whether to save the plot. By default the plot is not saved.
        filename : str, optional  (default="")
            Choose a filename for the plot that you want to save. This includes
            the file extension.
        directory : str, optional  (default="")
            Select the directory in which you want the plot to be saved.

        Examples
        --------
        Plot the first dissimilar frame for each frame in the dataset used as
        the reference frame.

        >>>  scat_obj.plot_first_n_diff_frames()

        Plot the first of 3 consecutive dissimilar frames for each frame in the
        dataset used as the reference frame (as was the case described in
        Brooks-Bartlett et al. 2016 - Development of tools to automate
        quantitative analysis of radiation damage in SAXS experiments).

        >>>  scat_obj.plot_first_n_diff_frames(n=3)

        Save the plot in the current working directory without displaying it.

        >>>  scat_obj.plot_first_n_diff_frames(display=False, save=True, filename="MyPlot.png")
        """
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                          SET PLOT PARAMS                        #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.PLOT_NUM += 1

        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                            PLOT CURVE                           #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        num_frames = self.I.shape[1]
        diff_frames_list = np.zeros(num_frames)
        frames = np.linspace(1, num_frames, num_frames)
        for i in range(0, num_frames):
            frame = i+1
            diff_frames_list[i] = self.find_first_n_diff_frames(n=n, frame=frame, P_threshold=P_threshold, P_type=P_type)

        plt.figure(self.PLOT_NUM)
        plt.plot(frames, diff_frames_list)
        plt.xlabel("Reference Frame Number")
        if n == 1:
            plt.ylabel("First Dissimilar Frame")
        else:
            plt.ylabel("First of {} Consecutive Dissimilar Frames".format(n))
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        #                       SAVE AND/OR DISPLAY                       #
        # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if save and filename:
            if directory:
                plot_path = "{}/{}".format(directory, filename)
            else:
                plot_path = filename
            plt.savefig(plot_path)
        elif save and not filename:
            print("********************** ERROR ***************************")
            print("COULD NOT SAVE PLOT")
            print("No filename specified. Please specify a filename if you")
            print("would like to save the plot.")
        if display:
            plt.show()

# --------------------------------------------------------------------------- #
#                               FUNCTIONS                                     #
# --------------------------------------------------------------------------- #


def run_system_command(command_string):
    """Function used to run the system command and return the log"""
    process = subprocess.Popen(command_string, stdout=subprocess.PIPE,
                               shell=True)  # Run system command
    output = process.communicate()  # Get the log.
    return output[0].decode()  # return the log file


def parse_saxs_dat_file(filename, data_type, array_size=-1):
    """Parse .dat file containing SAXS data."""
    if array_size == -1:
        array_size_set = False
        array_size = 10000
    else:
        array_size_set = True

    if data_type == "scattering angle":
        data_column = 0
    elif data_type == "intensity":
        data_column = 1
    elif data_column == "sigma":
        data_column = 2
        print("******WARNING*******")
        print("Data column not specified correctly. Returning the intensity column by default.")
        print("Please enter 'scattering angle', 'intensity' or 'sigma' as the data_type.")
        data_column = 1
    data_array = -1 * np.ones([array_size])
    prog = re.compile("(?:\d+\.)?\d+[e][-\+]\d+")
    file_obj = open(filename)
    counter = 0
    for line in file_obj:
        split_line = line.strip().split()
        if prog.match(line.strip()) and len(split_line) == 3:
            data_array[counter] = float(split_line[data_column])
            counter += 1
    if not array_size_set:
        data_array = data_array[data_array >= 0]
    file_obj.close()
    return data_array


def crop_saxs_files(file_list, temp_dir_name, temp_file_prefix, first, last, smin, smax):
    """Crop SAXS files.

    Uses DATCROP from the ATSAS suite of programs to crop the SAXS Data.
    """
    for index, filename in enumerate(file_list):
        if first > 0:
            if last > 0:
                datcrop_command = "datcrop {} --first {} --last {} -o {}/{}{}.dat".format(filename, first, last, temp_dir_name, temp_file_prefix, index)
            elif smax > 0:
                datcrop_command = "datcrop {} --first {} --smax {} -o {}/{}{}.dat".format(filename, first, smax, temp_dir_name, "temp", index)
            else:
                datcrop_command = "datcrop {} --first {} -o {}/{}{}.dat".format(filename, first, temp_dir_name, temp_file_prefix, index)
        elif smin > 0:
            if last > 0:
                datcrop_command = "datcrop {} --smin {} --last {} -o {}/{}{}.dat".format(filename, smin, last, temp_dir_name, temp_file_prefix, index)
            elif smax > 0:
                datcrop_command = "datcrop {} --smin {} --smax {} -o {}/{}{}.dat".format(filename, smin, smax, temp_dir_name,temp_file_prefix, index)
            else:
                datcrop_command = "datcrop {} --smin {} -o {}/{}{}.dat".format(filename, smin, temp_dir_name, temp_file_prefix, index)
        elif last > 0:
            if first > 0:
                datcrop_command = "datcrop {} --last {} --first {} -o {}/{}{}.dat".format(filename, last, first, temp_dir_name, temp_file_prefix, index)
            elif smin > 0:
                datcrop_command = "datcrop {} --last {} --smin {} -o {}/{}{}.dat".format(filename, last, smin, temp_dir_name, temp_file_prefix, index)
            else:
                datcrop_command = "datcrop {} --last {} -o {}/{}{}.dat".format(filename, last, temp_dir_name, temp_file_prefix, index)
        elif smax > 0:
            if first > 0:
                datcrop_command = "datcrop {} --smax {} --first {} -o {}/{}{}.dat".format(filename, smax, first, temp_file_prefix, index)
            elif smin > 0:
                datcrop_command = "datcrop {} --smax {} --smin {} -o {}/{}{}.dat".format(filename, smax, smin, temp_dir_name, temp_file_prefix, index)
            else:
                datcrop_command = "datcrop {} --smax {} -o {}/{}{}.dat".format(filename, smax, temp_dir_name, temp_file_prefix, index)
        run_system_command(datcrop_command)


def on_draw(event):
    """Auto-wraps all text objects in a figure at draw-time."""
    import matplotlib as mpl
    fig = event.canvas.figure

    # Cycle through all artists in all the axes in the figure
    for ax in fig.axes:
        for artist in ax.get_children():
            # If it's a text artist, wrap it...
            if isinstance(artist, mpl.text.Text):
                autowrap_text(artist, event.renderer)

    # Temporarily disconnect any callbacks to the draw event...
    # (To avoid recursion)
    func_handles = fig.canvas.callbacks.callbacks[event.name]
    fig.canvas.callbacks.callbacks[event.name] = {}
    # Re-draw the figure..
    fig.canvas.draw()
    # Reset the draw event callbacks
    fig.canvas.callbacks.callbacks[event.name] = func_handles


def autowrap_text(textobj, renderer):
    """Wraps the given matplotlib text object so that it exceed the boundaries
    of the axis it is plotted in."""
    import textwrap
    # Get the starting position of the text in pixels...
    x0, y0 = textobj.get_transform().transform(textobj.get_position())
    # Get the extents of the current axis in pixels...
    clip = textobj.axes.get_window_extent()
    # Set the text to rotate about the left edge (doesn't make sense otherwise)
    textobj.set_rotation_mode('anchor')

    # Get the amount of space in the direction of rotation to the left and
    # right of x0, y0 (left and right are relative to the rotation, as well)
    rotation = textobj.get_rotation()
    right_space = min_dist_inside((x0, y0), rotation, clip)
    left_space = min_dist_inside((x0, y0), rotation - 180, clip)

    # Use either the left or right distance depending on the horiz alignment.
    alignment = textobj.get_horizontalalignment()
    if alignment == 'left':
        new_width = right_space
    elif alignment == 'right':
        new_width = left_space
    else:
        new_width = 2 * min(left_space, right_space)

    # Estimate the width of the new size in characters...
    aspect_ratio = 0.5  # This varies with the font!!
    fontsize = textobj.get_size()
    pixels_per_char = aspect_ratio * renderer.points_to_pixels(fontsize)

    # If wrap_width is < 1, just make it 1 character
    wrap_width = max(1, new_width // pixels_per_char)
    try:
        wrapped_text = textwrap.fill(textobj.get_text(), wrap_width)
    except TypeError:
        # This appears to be a single word
        wrapped_text = textobj.get_text()
    textobj.set_text(wrapped_text)


def min_dist_inside(point, rotation, box):
    """Gets the space in a given direction from "point" to the boundaries of
    "box" (where box is an object with x0, y0, x1, & y1 attributes, point is a
    tuple of x,y, and rotation is the angle in degrees)"""
    from math import sin, cos, radians
    x0, y0 = point
    rotation = radians(rotation)
    distances = []
    threshold = 0.0001
    if cos(rotation) > threshold:
        # Intersects the right axis
        distances.append((box.x1 - x0) / cos(rotation))
    if cos(rotation) < -threshold:
        # Intersects the left axis
        distances.append((box.x0 - x0) / cos(rotation))
    if sin(rotation) > threshold:
        # Intersects the top axis
        distances.append((box.y1 - y0) / sin(rotation))
    if sin(rotation) < -threshold:
        # Intersects the bottom axis
        distances.append((box.y0 - y0) / sin(rotation))
    return min(distances)



"""
    UV.Foldedness.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from bisect import bisect_right

class Foldedness:
    def __init__(self, wv):
        self.wv = wv
        self.indeces = [bisect_right(wv, w) for w in self.get_wavelengths()]

    def get_wavelengths(self):
        return [280, 275, 258]

    def get_plotcolors(self):
        return ["red", "cyan", "gray"]

    def compute(self, pv):
        i = self.indeces
        return pv[i[0]]/pv[i[1]] + pv[i[0]]/pv[i[2]]

if __name__ == '__main__':
    import numpy as np
    wv = np.arange(250,300)
    f = Foldedness(wv)
    print(f.compute(np.arange(50)))



# coding: utf-8
"""

    ファイル名：   TkCustomWidgets.py

    処理内容：

        FileNodeEntry ----- 上位クラス
        FolderEntry ------- Drag&drop 機能および選択ボタン付きの Folder 入力用 Entry
        FileEntry   ------- Drag&drop 機能および選択ボタン付きの File 入力用 Entry

    Copyright (c) 2017-2021, Masatsuyo Takahashi

"""
import os

from molass_legacy.KekLib.OurTkinter import Tk, FileDialog, is_empty_val
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox    as MessageBox
from molass_legacy.KekLib.TkSupplements import SlimButton
from TkSizableWidgets import SizableEntry

class FileNodeEntry( Tk.Frame ):
    def __init__( self, parent, **kw ):

        sizable = kw.get( 'sizable' )
        if sizable is None:
            sizable = False
        else:
            del kw[ 'sizable' ]

        slimbutton = kw.get( 'slimbutton' )
        if slimbutton is None:
            slimbutton = False
        else:
            del kw[ 'slimbutton' ]

        Tk.Frame.__init__( self, parent )
        self.parent = parent
        var = kw.get( 'textvariable' )
        if var is None:
            var = Tk.StringVar()
            kw['textvariable'] = var
        self.variable = var
        self.on_entry_cb = kw.pop('on_entry_cb', None)

        if sizable:
            self.entry  = SizableEntry( self, **kw )
        else:
            self.entry  = Tk.Entry( self, **kw )

        if slimbutton:
            # self.button = SlimButton( self, text='...', command=self.on_button_click )
            self.button = SlimButton( self, text='...', command=self.on_button_click, width=18, height=22 )
        else:
            self.button = Tk.Button( self, text='...', command=self.on_button_click )

        self.entry.pack( side= Tk.LEFT, fill=Tk.BOTH, expand=1 )
        self.button.pack( side=Tk.LEFT )
        self.add_dnd_bind()

    def delete( self, *args ):
        self.entry.delete( *args )

    def insert( self, *args ):
        self.entry.insert( *args )

    def config( self, *args, **kwargs ):
        for w in [ self.entry, self.button ]:
            w.config( *args, **kwargs )

    def add_dnd_bind( self ):
        self.entry.register_drop_target("*")

        def dnd_handler( event ):
            event.widget.delete( 0, Tk.END )
            event.widget.insert( 0, event.data )
            self.on_entry()

        self.entry.bind("<<Drop>>", dnd_handler)

    def on_button_click( self ):
        f = self.select()
        if not f:
            return

        self.variable.set( f )
        self.check()
        self.on_entry()

    def select( self ):
        # must be overridden
        assert( False )

    def check( self ):
        # must be overridden
        assert( False )

    def on_entry( self, *args, **kwargs ):
        ok = self.check( *args )
        if not ok:
            return

        if self.on_entry_cb is not None:
            self.on_entry_cb( *args, **kwargs )

    def set_error( self ):
        self.entry.config( fg='red' )
        self.entry.focus_force()

    def bind(self, *args):
        self.entry.bind(*args)

class FolderEntry( FileNodeEntry ):
    def __init__( self, parent, **kw ):
        FileNodeEntry.__init__( self, parent, **kw )

    def select( self ):
        path = self.variable.get()
        dir_ = os.path.dirname( path ).replace( '/', '\\' )
        # print 'dir_=', dir_
        f = FileDialog.askdirectory( initialdir=dir_, parent=self.parent )
        return f

    def check( self, *args ):
        f = self.variable.get()
        ret = False
        if is_empty_val( f ):
            self.set_error()
            MessageBox.showerror( "Folder Input Error", "This folder input is required.", parent=self.parent )
        elif not os.path.exists( f ):
            self.set_error()
            MessageBox.showerror( "Folder Input Error", "'%s' does not exist." % f, parent=self.parent )
        elif not os.path.isdir( f ):
            self.set_error()
            MessageBox.showerror( "Folder Input Error", "'%s' is not a folder." % f, parent=self.parent )
        else:
            self.entry.config( fg='black' )
            ret = True

        if not ret:
            self.entry.focus_force()

        return ret

class FileEntry( FileNodeEntry ):
    def __init__( self, parent, **kw ):
        self.mode = kw.pop( 'mode', None )
        if self.mode is None:
            self.mode = 'r'

        FileNodeEntry.__init__( self, parent, **kw )

    def select( self ):
        path = self.variable.get()
        dir_ = os.path.dirname( path ).replace( '/', '\\' )
        # print 'dir_=', dir_
        if self.mode == 'w':
            f = FileDialog.asksaveasfilename ( initialdir=dir_, parent=self.parent )
        else:
            f = FileDialog.askopenfilename( initialdir=dir_, parent=self.parent )
        return f

    def check( self, *args ):
        f = self.variable.get()
        print( 'check: f=', f )
        ret = False
        if is_empty_val( f ):
            self.set_error()
            MessageBox.showerror( "File Input Error", "This file input is required.", parent=self.parent )
        elif not os.path.exists( f ):
            self.set_error()
            MessageBox.showerror( "File Input Error", "'%s' does not exist." % f, parent=self.parent )
        elif not os.path.isfile( f ):
            self.set_error()
            MessageBox.showerror( "File Input Error", "'%s' is not a file." % f, parent=self.parent )
        else:
            self.entry.config( fg='black' )
            ret = True

        if not ret:
            self.entry.focus_force()

        return ret

if __name__ == "__main__":
    root = Tk.Tk()
    var = Tk.StringVar()
    entry   = FolderEntry( root, textvariable=var, width=60 )
    # entry   = FileEntry( root, textvariable=var, width=60 )
    entry.pack()
    root.mainloop()



"""

    ThreeDimViewer.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF

"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from DataUtils import get_in_folder
from MatrixData import simple_plot_3d

class ThreeDimViewer(Dialog):
    def __init__(self, parent, md):
        self.md = md
        self.popup_menu = None
        Dialog.__init__(self, parent, "3D Viewer", visible=False)

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig)
        Dialog.cancel(self)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tbframe = Tk.Frame(body_frame)
        tbframe.pack(fill=Tk.X, expand=1)

        md = self.md
        self.fig = fig = plt.figure(figsize=(21,7))

        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()

        in_folder = get_in_folder()
        fig.suptitle(in_folder, fontsize=20)
        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132, projection='3d')
        ax3 = fig.add_subplot(133)
        self.axes = [ax1, ax2, ax3]
        md.uv.plot(ax=ax1)
        md.xr.plot(ax=ax2)
        md.plot_pre_sync(ax3)

        """
            show guide for how to zoom
            learned at https://pythonprogramming.net/3d-graphs-matplotlib-tutorial/
        """
        axg = fig.add_axes([0,0,0.66,0.05])
        axg.set_axis_off()
        axg.text(0.5, 0.3, "Drag with the right mouse button to zoom in or out in 3d plots.", fontsize=16, alpha=0.3, ha='center')

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )
        self.mpl_canvas.draw()

    def show(self):
        self._show()

    def save_the_figure(self, folder=None, file=None):
        import os
        from molass_legacy._MOLASS.SerialSettings import get_setting
        if folder is None:
            folder = get_setting('analysis_folder')
        if file is None:
            file = get_setting('analysis_name')
        path = os.path.join(folder, file)
        self.fig.savefig(path)

    def on_button_press(self, event):
        if event.button == 3:
            self.on_right_button_press(event)
            return

    def on_right_button_press(self, event):
        if event.inaxes != self.axes[1]:
            return

        if self.popup_menu is None:
            self.create_popup_menu()

        rootx = self.winfo_rootx()
        rooty = self.winfo_rooty()
        w, h, x, y = split_geometry(self.mpl_canvas_widget.winfo_geometry())
        self.popup_menu.post(rootx + int(event.x), rooty + h - int(event.y))

    def create_popup_menu(self):
        self.popup_menu = Tk.Menu(self, tearoff=0)
        self.popup_menu.add_command(label='Show error data', command=self.show_error_data)

    def show_error_data(self):
        dialog = ThreeDimErrorViewer(self, self.md.xr)
        dialog.show()

class ThreeDimErrorViewer(Dialog):
    def __init__(self, parent, xr):
        self.xr = xr
        Dialog.__init__(self, parent, "3D Error View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tbframe = Tk.Frame(body_frame)
        tbframe.pack(fill=Tk.X, expand=1)

        xr = self.xr

        self.fig = fig = plt.figure(figsize=(14,7))

        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()

        in_folder = get_in_folder()
        fig.suptitle(in_folder, fontsize=20)
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122, projection='3d')

        ax1.set_title("Xray Scattering Error (Linear)", fontsize=16, y=1.08)
        simple_plot_3d(ax1, xr.error.data, x=xr.vector)

        ax2.set_title("Xray Scattering Error (Log10)", fontsize=16, y=1.08)
        simple_plot_3d(ax2, np.log10(xr.error.data), x=xr.vector)

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)

        self.mpl_canvas.draw()
